@@ -19,6 +19,9 @@
 
 #include "common.h"
 
+#include <libplacebo/colorspace.h>
+#include <libplacebo/tone_mapping.h>
+
 bool pl_color_system_is_ycbcr_like(enum pl_color_system sys)
 {
     switch (sys) {
@@ -202,6 +205,8 @@ bool pl_color_primaries_is_wide_gamut(enum pl_color_primaries prim)
     case PL_COLOR_PRIM_V_GAMUT:
     case PL_COLOR_PRIM_S_GAMUT:
     case PL_COLOR_PRIM_FILM_C:
+    case PL_COLOR_PRIM_ACES_AP0:
+    case PL_COLOR_PRIM_ACES_AP1:
         return true;
     case PL_COLOR_PRIM_COUNT: break;
     }
@@ -242,6 +247,7 @@ float pl_color_transfer_nominal_peak(enum pl_color_transfer trc)
     case PL_COLOR_TRC_GAMMA26:
     case PL_COLOR_TRC_GAMMA28:
     case PL_COLOR_TRC_PRO_PHOTO:
+    case PL_COLOR_TRC_ST428:
         return 1.0;
     case PL_COLOR_TRC_PQ:       return 10000.0 / PL_COLOR_SDR_WHITE;
     case PL_COLOR_TRC_HLG:      return 12.0 / PL_COLOR_SDR_WHITE_HLG;
@@ -284,6 +290,73 @@ const struct pl_hdr_metadata pl_hdr_metadata_hdr10 ={
     .max_fall = 0, // unknown
 };
 
+static const float PQ_M1 = 2610./4096 * 1./4,
+                   PQ_M2 = 2523./4096 * 128,
+                   PQ_C1 = 3424./4096,
+                   PQ_C2 = 2413./4096 * 32,
+                   PQ_C3 = 2392./4096 * 32;
+
+float pl_hdr_rescale(enum pl_hdr_scaling from, enum pl_hdr_scaling to, float x)
+{
+    if (from == to)
+        return x;
+    if (!x) // micro-optimization for common value
+        return x;
+
+    // Convert input to PL_SCALE_RELATIVE
+    switch (from) {
+    case PL_HDR_PQ:
+        x = powf(x, 1.0f / PQ_M2);
+        x = fmaxf(x - PQ_C1, 0.0f) / (PQ_C2 - PQ_C3 * x);
+        x = powf(x, 1.0f / PQ_M1);
+        x *= 10000.0f;
+        // fall through
+    case PL_HDR_NITS:
+        x /= PL_COLOR_SDR_WHITE;
+        // fall through
+    case PL_HDR_NORM:
+        goto output;
+    case PL_HDR_SQRT:
+        x *= x;
+        goto output;
+    case PL_HDR_SCALING_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+
+output:
+    // Convert PL_SCALE_RELATIVE to output
+    switch (to) {
+    case PL_HDR_NORM:
+        return x;
+    case PL_HDR_SQRT:
+        return sqrtf(x);
+    case PL_HDR_NITS:
+        return x * PL_COLOR_SDR_WHITE;
+    case PL_HDR_PQ:
+        x *= PL_COLOR_SDR_WHITE / 10000.0f;
+        x = powf(x, PQ_M1);
+        x = (PQ_C1 + PQ_C2 * x) / (1.0f + PQ_C3 * x);
+        x = powf(x, PQ_M2);
+        return x;
+    case PL_HDR_SCALING_COUNT:
+        break;
+    }
+
+    pl_unreachable();
+}
+
+static inline bool pl_hdr_bezier_equal(const struct pl_hdr_bezier *a,
+                                       const struct pl_hdr_bezier *b)
+{
+    return a->target_luma == b->target_luma &&
+           a->knee_x      == b->knee_x &&
+           a->knee_y      == b->knee_y &&
+           a->num_anchors == b->num_anchors &&
+           !memcmp(a->anchors, b->anchors, sizeof(a->anchors[0]) * a->num_anchors);
+}
+
 bool pl_hdr_metadata_equal(const struct pl_hdr_metadata *a,
                            const struct pl_hdr_metadata *b)
 {
@@ -291,7 +364,14 @@ bool pl_hdr_metadata_equal(const struct pl_hdr_metadata *a,
            a->min_luma == b->min_luma &&
            a->max_luma == b->max_luma &&
            a->max_cll  == b->max_cll  &&
-           a->max_fall == b->max_fall;
+           a->max_fall == b->max_fall &&
+           a->scene_max[0] == b->scene_max[0] &&
+           a->scene_max[1] == b->scene_max[1] &&
+           a->scene_max[2] == b->scene_max[2] &&
+           a->scene_avg == b->scene_avg &&
+           pl_hdr_bezier_equal(&a->ootf, &b->ootf) &&
+           a->max_pq_y == b->max_pq_y &&
+           a->avg_pq_y == b->avg_pq_y;
 }
 
 void pl_hdr_metadata_merge(struct pl_hdr_metadata *orig,
@@ -306,6 +386,37 @@ void pl_hdr_metadata_merge(struct pl_hdr_metadata *orig,
         orig->max_cll = update->max_cll;
     if (!orig->max_fall)
         orig->max_fall = update->max_fall;
+    if (!orig->scene_max[1])
+        memcpy(orig->scene_max, update->scene_max, sizeof(orig->scene_max));
+    if (!orig->scene_avg)
+        orig->scene_avg = update->scene_avg;
+    if (!orig->ootf.target_luma)
+        orig->ootf = update->ootf;
+    if (!orig->max_pq_y)
+        orig->max_pq_y = update->max_pq_y;
+    if (!orig->avg_pq_y)
+        orig->avg_pq_y = update->avg_pq_y;
+}
+
+bool pl_hdr_metadata_contains(const struct pl_hdr_metadata *data,
+                              enum pl_hdr_metadata_type type)
+{
+    bool has_hdr10 = data->min_luma && data->max_luma;
+    bool has_hdr10plus = data->scene_avg && (data->scene_max[0] ||
+                                             data->scene_max[1] ||
+                                             data->scene_max[2]);
+    bool has_cie_y = data->max_pq_y && data->avg_pq_y;
+
+    switch (type) {
+    case PL_HDR_METADATA_NONE:          return true;
+    case PL_HDR_METADATA_ANY:           return has_hdr10 || has_hdr10plus || has_cie_y;
+    case PL_HDR_METADATA_HDR10:         return has_hdr10;
+    case PL_HDR_METADATA_HDR10PLUS:     return has_hdr10plus;
+    case PL_HDR_METADATA_CIE_Y:         return has_cie_y;
+    case PL_HDR_METADATA_TYPE_COUNT:    break;
+    }
+
+    pl_unreachable();
 }
 
 const struct pl_color_space pl_color_space_unknown = {0};
@@ -338,7 +449,6 @@ const struct pl_color_space pl_color_space_monitor = {
 bool pl_color_space_is_hdr(const struct pl_color_space *csp)
 {
     return csp->hdr.max_luma > PL_COLOR_SDR_WHITE ||
-           csp->sig_scale > 1 ||
            pl_color_transfer_is_hdr(csp->transfer);
 }
 
@@ -355,6 +465,7 @@ bool pl_color_space_is_black_scaled(const struct pl_color_space *csp)
     case PL_COLOR_TRC_GAMMA26:
     case PL_COLOR_TRC_GAMMA28:
     case PL_COLOR_TRC_PRO_PHOTO:
+    case PL_COLOR_TRC_ST428:
     case PL_COLOR_TRC_HLG:
         return true;
 
@@ -389,36 +500,51 @@ bool pl_color_space_equal(const struct pl_color_space *c1,
            pl_hdr_metadata_equal(&c1->hdr, &c2->hdr);
 }
 
-void pl_color_space_infer(struct pl_color_space *space)
+// Estimates luminance from maxRGB by looking at how monochromatic MaxSCL is
+static void luma_from_maxrgb(const struct pl_color_space *csp,
+                             enum pl_hdr_scaling scaling,
+                             float *out_max, float *out_avg)
 {
-    if (!space->primaries)
-        space->primaries = PL_COLOR_PRIM_BT_709;
-    if (!space->transfer)
-        space->transfer = PL_COLOR_TRC_BT_1886;
+    const float maxscl = PL_MAX3(csp->hdr.scene_max[0],
+                                 csp->hdr.scene_max[1],
+                                 csp->hdr.scene_max[2]);
+    if (!maxscl)
+        return;
 
-    // Backwards-compatibility with deprecated fields
-    if (space->sig_peak) {
-        space->hdr.max_luma = space->sig_peak * PL_COLOR_SDR_WHITE;
-        space->sig_peak = 0;
-    }
-    if (space->sig_floor) {
-        space->hdr.min_luma = space->sig_floor * PL_COLOR_SDR_WHITE;
-        space->sig_floor = 0;
-    }
+    struct pl_raw_primaries prim = csp->hdr.prim;
+    pl_raw_primaries_merge(&prim, pl_raw_primaries_get(csp->primaries));
+    const struct pl_matrix3x3 rgb2xyz = pl_get_rgb2xyz_matrix(&prim);
 
-reinfer_peaks:
-    if (space->hdr.max_luma < 1 || space->hdr.max_luma > 10000) {
-        space->hdr.max_luma = pl_color_transfer_nominal_peak(space->transfer)
-                              * PL_COLOR_SDR_WHITE;
+    const float max_luma = rgb2xyz.m[1][0] * csp->hdr.scene_max[0] +
+                           rgb2xyz.m[1][1] * csp->hdr.scene_max[1] +
+                           rgb2xyz.m[1][2] * csp->hdr.scene_max[2];
+
+    const float coef = max_luma / maxscl;
+    *out_max = pl_hdr_rescale(PL_HDR_NITS, scaling, max_luma);
+    *out_avg = pl_hdr_rescale(PL_HDR_NITS, scaling, coef * csp->hdr.scene_avg);
+}
+
+void pl_color_space_nominal_luma_ex(const struct pl_nominal_luma_params *params)
+{
+    if (!params || (!params->out_min && !params->out_max && !params->out_avg))
+        return;
+
+    const struct pl_color_space *csp = params->color;
+    const enum pl_hdr_scaling scaling = params->scaling;
+
+    // Baseline/fallback metadata, inferred entirely from the colorspace
+    // description and built-in default assumptions
+    if (params->metadata == PL_HDR_METADATA_NONE) {
+        float min_luma, max_luma; // in nits
+        max_luma = pl_color_transfer_nominal_peak(csp->transfer) *
+                   PL_COLOR_SDR_WHITE;
 
         // Exception: For HLG content, we want to infer a value of 1000 cd/m²,
         // a value which is considered the "reference" HLG display.
-        if (space->transfer == PL_COLOR_TRC_HLG)
-            space->hdr.max_luma = 1000;
-    }
+        if (csp->transfer == PL_COLOR_TRC_HLG)
+            max_luma = 1000;
 
-    if (space->hdr.min_luma <= 0 || space->hdr.min_luma > 100) {
-        if (pl_color_transfer_is_hdr(space->transfer)) {
+        if (pl_color_transfer_is_hdr(csp->transfer)) {
             // Use a slightly nonzero black level, for the following reasons:
             // - 0 may be seen as 'missing/undefined' in HDR10 metadata structs
             //
@@ -429,29 +555,102 @@ reinfer_peaks:
             // - true infinite contrast does not exist in reality, even 1e-7
             //   is extremely generous considering typical viewing environments
             //   are not absolutely devoid of stray ambient light
-            space->hdr.min_luma = 1e-7;
+            min_luma = 1e-7;
         } else {
-            space->hdr.min_luma = space->hdr.max_luma / 1000; // Typical SDR contrast
+            min_luma = max_luma / 1000; // Typical SDR contrast
         }
+
+        if (params->out_min)
+            *params->out_min = pl_hdr_rescale(PL_HDR_NITS, scaling, min_luma);
+        if (params->out_max)
+            *params->out_max = pl_hdr_rescale(PL_HDR_NITS, scaling, max_luma);
+        if (params->out_avg)
+            *params->out_avg = 0;
+        return;
     }
 
-    pl_assert(space->hdr.min_luma && space->hdr.max_luma);
-    if (space->hdr.max_luma < space->hdr.min_luma) { // sanity
-        space->hdr.max_luma = space->hdr.min_luma = 0;
-        goto reinfer_peaks;
+    // Always take the minimum luminance from static mastering display
+    // metadata, because DV/HDR10+ don't encode this at all.
+    float min_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.min_luma);
+
+    // PQ is always scaled down to absolute black, ignoring HDR metadata
+    if (csp->transfer == PL_COLOR_TRC_PQ)
+        min_luma = 0;
+
+    float max_luma = 0, avg_luma = 0;
+    if (!params->metadata || params->metadata == PL_HDR_METADATA_HDR10) {
+        // This metadata is very often completely broken
+        if (csp->hdr.max_luma >= 100.0f)
+            max_luma = pl_hdr_rescale(PL_HDR_NITS, scaling, csp->hdr.max_luma);
     }
 
-    if (space->sig_scale && !pl_color_transfer_is_hdr(space->transfer)) {
-        space->hdr.max_luma *= space->sig_scale;
-        space->hdr.min_luma *= space->sig_scale;
-        space->sig_scale = 0;
+    if ((!params->metadata || params->metadata == PL_HDR_METADATA_HDR10PLUS) &&
+        pl_hdr_metadata_contains(&csp->hdr, PL_HDR_METADATA_HDR10PLUS))
+    {
+        luma_from_maxrgb(csp, scaling, &max_luma, &avg_luma);
     }
 
-    if (!pl_primaries_valid(&space->hdr.prim))
-        memset(&space->hdr.prim, 0, sizeof(space->hdr.prim));
+    if ((!params->metadata || params->metadata == PL_HDR_METADATA_CIE_Y) &&
+        pl_hdr_metadata_contains(&csp->hdr, PL_HDR_METADATA_CIE_Y))
+    {
+        max_luma = pl_hdr_rescale(PL_HDR_PQ, scaling, csp->hdr.max_pq_y);
+        avg_luma = pl_hdr_rescale(PL_HDR_PQ, scaling, csp->hdr.avg_pq_y);
+    }
+
+    if (max_luma && min_luma > max_luma) // sanity
+        min_luma = max_luma = 0;
+
+    // Make sure this metadata is complete by inferring with fixed defaults
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_NONE,
+        .scaling    = scaling,
+        .out_min    = min_luma ? NULL : &min_luma,
+        .out_max    = max_luma ? NULL : &max_luma,
+    ));
+
+    if (avg_luma)
+        avg_luma = PL_CLAMP(avg_luma, min_luma, max_luma); // sanity
+
+    if (params->out_min)
+        *params->out_min = min_luma;
+    if (params->out_max)
+        *params->out_max = max_luma;
+    if (params->out_avg)
+        *params->out_avg = avg_luma;
+}
+
+void pl_color_space_nominal_luma(const struct pl_color_space *csp,
+                                 float *out_min, float *out_max)
+{
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = csp,
+        .metadata   = PL_HDR_METADATA_ANY,
+        .scaling    = PL_HDR_NORM,
+        .out_min    = out_min,
+        .out_max    = out_max,
+    ));
+}
+
+void pl_color_space_infer(struct pl_color_space *space)
+{
+    if (!space->primaries)
+        space->primaries = PL_COLOR_PRIM_BT_709;
+    if (!space->transfer)
+        space->transfer = PL_COLOR_TRC_BT_1886;
+
+    // Sanitize the static HDR metadata
+    pl_color_space_nominal_luma_ex(pl_nominal_luma_params(
+        .color      = space,
+        .metadata   = PL_HDR_METADATA_HDR10,
+        .scaling    = PL_HDR_NITS,
+        .out_min    = &space->hdr.min_luma,
+        .out_max    = &space->hdr.max_luma,
+    ));
 
     // Default the signal color space based on the nominal raw primaries
-    pl_raw_primaries_merge(&space->hdr.prim, pl_raw_primaries_get(space->primaries));
+    if (!pl_primaries_valid(&space->hdr.prim))
+        space->hdr.prim = *pl_raw_primaries_get(space->primaries);
 }
 
 static void infer_both_ref(struct pl_color_space *space,
@@ -468,12 +667,39 @@ static void infer_both_ref(struct pl_color_space *space,
     }
 
     if (!space->transfer) {
-        if (pl_color_transfer_is_hdr(ref->transfer)) {
-            space->transfer = PL_COLOR_TRC_BT_1886;
-        } else if (ref->transfer == PL_COLOR_TRC_LINEAR) {
-            space->transfer = PL_COLOR_TRC_GAMMA22;
-        } else {
+        switch (ref->transfer) {
+        case PL_COLOR_TRC_UNKNOWN:
+        case PL_COLOR_TRC_COUNT:
+            pl_unreachable();
+        case PL_COLOR_TRC_BT_1886:
+        case PL_COLOR_TRC_SRGB:
+        case PL_COLOR_TRC_GAMMA22:
+            // Re-use input transfer curve to avoid small adaptations
             space->transfer = ref->transfer;
+            break;
+        case PL_COLOR_TRC_PQ:
+        case PL_COLOR_TRC_HLG:
+        case PL_COLOR_TRC_V_LOG:
+        case PL_COLOR_TRC_S_LOG1:
+        case PL_COLOR_TRC_S_LOG2:
+            // Pick BT.1886 model because it models SDR contrast accurately,
+            // and we need contrast information for tone mapping
+            space->transfer = PL_COLOR_TRC_BT_1886;
+            break;
+        case PL_COLOR_TRC_PRO_PHOTO:
+            // ProPhotoRGB and sRGB are both piecewise with linear slope
+            space->transfer = PL_COLOR_TRC_SRGB;
+            break;
+        case PL_COLOR_TRC_LINEAR:
+        case PL_COLOR_TRC_GAMMA18:
+        case PL_COLOR_TRC_GAMMA20:
+        case PL_COLOR_TRC_GAMMA24:
+        case PL_COLOR_TRC_GAMMA26:
+        case PL_COLOR_TRC_GAMMA28:
+        case PL_COLOR_TRC_ST428:
+            // Pick pure power output curve to avoid introducing black crush
+            space->transfer = PL_COLOR_TRC_GAMMA22;
+            break;
         }
     }
 
@@ -494,17 +720,6 @@ void pl_color_space_infer_map(struct pl_color_space *src,
 {
     bool unknown_contrast = !src->hdr.min_luma;
 
-    // PQ is always scaled down to absolute black, regardless of what the
-    // metadata says. So treat it as such. We do this here, rather than in e.g.
-    // pl_color_space_infer, because the latter is also used to sanitize values
-    // in cases where we don't want to override mastering metadata.
-    // Additionally, doing it here allows e.g. correct propagation of black
-    // levels for BT.1886 -> PQ expansion in the step further below.
-    if (src->transfer == PL_COLOR_TRC_PQ)
-        src->hdr.min_luma = 0;
-    if (dst->transfer == PL_COLOR_TRC_PQ)
-        dst->hdr.min_luma = 0;
-
     infer_both_ref(dst, src);
 
     // If the src has an unspecified SDR-like gamma, default it to match the
@@ -628,11 +843,11 @@ const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim
     */
 
     // CIE standard illuminant series
-#define CIE_D50 {0.34577, 0.35850}
-#define CIE_D65 {0.31271, 0.32902}
-#define CIE_DCI {0.31400, 0.35100}
-#define CIE_C   {0.31006, 0.31616}
+#define CIE_D50 {0.3457, 0.3585}
+#define CIE_D65 {0.3127, 0.3290}
+#define CIE_C   {0.3100, 0.3160}
 #define CIE_E   {1.0/3.0, 1.0/3.0}
+#define DCI     {0.3140, 0.3510}
 
     static const struct pl_raw_primaries primaries[] = {
         [PL_COLOR_PRIM_BT_470M] = {
@@ -695,7 +910,7 @@ const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim
             .red   = {0.680, 0.320},
             .green = {0.265, 0.690},
             .blue  = {0.150, 0.060},
-            .white = CIE_DCI,
+            .white = DCI,
         },
         [PL_COLOR_PRIM_DISPLAY_P3] = {
             .red   = {0.680, 0.320},
@@ -730,6 +945,19 @@ const struct pl_raw_primaries *pl_raw_primaries_get(enum pl_color_primaries prim
             .blue  = {0.155, 0.077},
             .white = CIE_D65,
         },
+    // From Wikipedia
+        [PL_COLOR_PRIM_ACES_AP0] = {
+            .red   = {0.7347, 0.2653},
+            .green = {0.0000, 1.0000},
+            .blue  = {0.0001, -0.0770},
+            .white = {0.32168, 0.33767},
+        },
+        [PL_COLOR_PRIM_ACES_AP1] = {
+            .red   = {0.713, 0.293},
+            .green = {0.165, 0.830},
+            .blue  = {0.128, 0.044},
+            .white = {0.32168, 0.33767},
+        },
     };
 
     // This is the default assumption if no colorspace information could
@@ -1166,8 +1394,11 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
         break;
     case PL_COLOR_SYSTEM_XYZ: {
         // For lack of anything saner to do, just assume the caller wants
-        // BT.709 primaries, which is a reasonable assumption.
-        m = pl_get_xyz2rgb_matrix(pl_raw_primaries_get(PL_COLOR_PRIM_BT_709));
+        // DCI-P3 primaries, which is a reasonable assumption.
+        const struct pl_raw_primaries *dst = pl_raw_primaries_get(PL_COLOR_PRIM_DCI_P3);
+        m = pl_get_xyz2rgb_matrix(dst);
+        // DCDM X'Y'Z' is expected to have equal energy white point (EG 432-1 Annex H)
+        apply_chromatic_adaptation((struct pl_cie_xy)CIE_E, dst->white, &m);
         break;
     }
     case PL_COLOR_SYSTEM_COUNT:
@@ -1230,6 +1461,7 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
     double mul[3]   = { ymul, ymul, ymul };
     double black[3] = { ymin, ymin, ymin };
 
+#ifdef PL_HAVE_DOVI
     if (repr->sys == PL_COLOR_SYSTEM_DOLBYVISION) {
         // The RPU matrix already includes levels normalization, but in this
         // case we also have to respect the signalled color offsets
@@ -1237,7 +1469,9 @@ struct pl_transform3x3 pl_color_repr_decode(struct pl_color_repr *repr,
             mul[i] = 1.0;
             black[i] = repr->dovi->nonlinear_offset[i] * scale;
         }
-    } else if (pl_color_system_is_ycbcr_like(repr->sys)) {
+    } else
+#endif
+    if (pl_color_system_is_ycbcr_like(repr->sys)) {
         mul[1]   = mul[2]   = cmul;
         black[1] = black[2] = cmid;
     }
