@@ -15,24 +15,14 @@
  * License along with libplacebo.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <ctype.h>
+
 #include "common.h"
 #include "utils.h"
 #include "gpu.h"
-#include "pl_thread.h"
 
 const struct pl_opengl_params pl_opengl_default_params = {0};
 
-struct priv {
-    struct pl_opengl_params params;
-    pl_log log;
-    bool is_debug;
-    bool is_debug_egl;
-
-    // For context locking
-    pl_mutex lock;
-    int count;
-};
-
 static void GLAPIENTRY debug_cb(GLenum source, GLenum type, GLuint id,
                                 GLenum severity, GLsizei length,
                                 const GLchar *message, const void *userParam)
@@ -55,11 +45,9 @@ static void GLAPIENTRY debug_cb(GLenum source, GLenum type, GLuint id,
 #endif
 }
 
-#ifdef EPOXY_HAS_EGL
-
-static void debug_cb_egl(EGLenum error, const char *command,
-                         EGLint messageType, EGLLabelKHR threadLabel,
-                         EGLLabelKHR objectLabel, const char *message)
+static void GLAPIENTRY debug_cb_egl(EGLenum error, const char *command,
+                                    EGLint messageType, EGLLabelKHR threadLabel,
+                                    EGLLabelKHR objectLabel, const char *message)
 {
     pl_log log = threadLabel;
     enum pl_log_level level = PL_LOG_ERR;
@@ -80,7 +68,8 @@ static void debug_cb_egl(EGLenum error, const char *command,
 #endif
 }
 
-#endif // EPOXY_HAS_EGL
+// Guards access to the (thread-unsafe) glad global EGL state
+static pl_static_mutex glad_egl_mutex = PL_STATIC_MUTEX_INITIALIZER;
 
 void pl_opengl_destroy(pl_opengl *ptr)
 {
@@ -88,31 +77,59 @@ void pl_opengl_destroy(pl_opengl *ptr)
     if (!pl_gl)
         return;
 
-    struct priv *p = PL_PRIV(pl_gl);
+    struct gl_ctx *p = PL_PRIV(pl_gl);
+    gl_funcs *gl = &p->func;
     if (!gl_make_current(pl_gl)) {
         PL_WARN(p, "Failed uninitializing OpenGL context, leaking resources!");
         return;
     }
 
     if (p->is_debug)
-        glDebugMessageCallback(NULL, NULL);
+        gl->DebugMessageCallback(NULL, NULL);
 
-#ifdef EPOXY_HAS_EGL
     if (p->is_debug_egl)
         eglDebugMessageControlKHR(NULL, NULL);
-#endif
 
     pl_gpu_destroy(pl_gl->gpu);
+
+#ifdef PL_HAVE_GL_PROC_ADDR
+    if (p->is_gles) {
+        gladLoaderUnloadGLES2Context(gl);
+    } else {
+        gladLoaderUnloadGLContext(gl);
+    }
+
+    bool used_loader = !p->params.get_proc_addr && !p->params.get_proc_addr_ex;
+    if (p->params.egl_display && used_loader) {
+        pl_static_mutex_lock(&glad_egl_mutex);
+        gladLoaderUnloadEGL();
+        pl_static_mutex_unlock(&glad_egl_mutex);
+    }
+#endif
+
     gl_release_current(pl_gl);
     pl_mutex_destroy(&p->lock);
     pl_free_ptr((void **) ptr);
+
+}
+
+typedef PL_ARRAY(const char *) ext_arr_t;
+static void add_exts_str(void *alloc, ext_arr_t *arr, const char *extstr)
+{
+    pl_str rest = pl_str_strip(pl_str0(pl_strdup0(alloc, pl_str0(extstr))));
+    while (rest.len) {
+        pl_str ext = pl_str_split_char(rest, ' ', &rest);
+        ext.buf[ext.len] = '\0'; // re-use separator for terminator
+        PL_ARRAY_APPEND(alloc, *arr, (char *) ext.buf);
+    }
 }
 
 pl_opengl pl_opengl_create(pl_log log, const struct pl_opengl_params *params)
 {
     params = PL_DEF(params, &pl_opengl_default_params);
-    struct pl_opengl *pl_gl = pl_zalloc_obj(NULL, pl_gl, struct priv);
-    struct priv *p = PL_PRIV(pl_gl);
+    struct pl_opengl_t *pl_gl = pl_zalloc_obj(NULL, pl_gl, struct gl_ctx);
+    struct gl_ctx *p = PL_PRIV(pl_gl);
+    gl_funcs *gl = &p->func;
     p->params = *params;
     p->log = log;
 
@@ -122,57 +139,139 @@ pl_opengl pl_opengl_create(pl_log log, const struct pl_opengl_params *params)
         return NULL;
     }
 
-    int ver = epoxy_gl_version();
-    if (!ver) {
+    bool ok;
+    if (params->get_proc_addr_ex) {
+        ok = gladLoadGLContextUserPtr(gl, params->get_proc_addr_ex, params->proc_ctx);
+    } else if (params->get_proc_addr) {
+        ok = gladLoadGLContext(gl, params->get_proc_addr);
+    } else {
+#ifdef PL_HAVE_GL_PROC_ADDR
+        ok = gladLoaderLoadGLContext(gl);
+#else
+        PL_FATAL(p, "No `glGetProcAddress` function provided, and libplacebo "
+                 "built without its built-in OpenGL loader!");
+        goto error;
+#endif
+    }
+
+    if (!ok) {
+        PL_INFO(p, "Failed loading core GL, retrying as GLES2...");
+    } else if (gl_is_gles(pl_gl)) {
+        PL_INFO(p, "GL context seems to be OpenGL ES, reloading as GLES2...");
+        ok = false;
+    }
+
+    if (!ok) {
+        memset(gl, 0, sizeof(*gl));
+        if (params->get_proc_addr_ex) {
+            ok = gladLoadGLES2ContextUserPtr(gl, params->get_proc_addr_ex, params->proc_ctx);
+        } else if (params->get_proc_addr) {
+            ok = gladLoadGLES2Context(gl, params->get_proc_addr);
+        } else {
+#ifdef PL_HAVE_GL_PROC_ADDR
+            ok = gladLoaderLoadGLES2Context(gl);
+#else
+            pl_unreachable();
+#endif
+        }
+        p->is_gles = ok;
+    }
+
+    if (!ok) {
+        PL_FATAL(p, "Failed to initialize OpenGL context - make sure a valid "
+                 "OpenGL context is bound to the current thread!");
+        goto error;
+    }
+
+    const char *version = (const char *) gl->GetString(GL_VERSION);
+    if (version) {
+        const char *ver = version;
+        while (!isdigit(*ver) && *ver != '\0')
+            ver++;
+        if (sscanf(ver, "%d.%d", &pl_gl->major, &pl_gl->minor) != 2) {
+            PL_FATAL(p, "Invalid GL_VERSION string: %s\n", version);
+            goto error;
+        }
+    }
+
+    if (!pl_gl->major) {
         PL_FATAL(p, "No OpenGL version detected - make sure an OpenGL context "
                  "is bound to the current thread!");
         goto error;
     }
 
     PL_INFO(p, "Detected OpenGL version strings:");
-    PL_INFO(p, "    GL_VERSION:  %s", (char *) glGetString(GL_VERSION));
-    PL_INFO(p, "    GL_VENDOR:   %s", (char *) glGetString(GL_VENDOR));
-    PL_INFO(p, "    GL_RENDERER: %s", (char *) glGetString(GL_RENDERER));
+    PL_INFO(p, "    GL_VERSION:  %s", version);
+    PL_INFO(p, "    GL_VENDOR:   %s", (char *) gl->GetString(GL_VENDOR));
+    PL_INFO(p, "    GL_RENDERER: %s", (char *) gl->GetString(GL_RENDERER));
 
-    if (pl_msg_test(log, PL_LOG_DEBUG)) {
-        if (ver >= 30) {
-            int num_exts = 0;
-            glGetIntegerv(GL_NUM_EXTENSIONS, &num_exts);
-            PL_DEBUG(p, "    GL_EXTENSIONS:");
-            for (int i = 0; i < num_exts; i++) {
-                const char *ext = (char *) glGetStringi(GL_EXTENSIONS, i);
-                PL_DEBUG(p, "        %s", ext);
-            }
-        } else {
-            PL_DEBUG(p, "    GL_EXTENSIONS: %s", (char *) glGetString(GL_EXTENSIONS));
-        }
-
-#ifdef EPOXY_HAS_EGL
-        if (params->egl_display) {
-            PL_DEBUG(p, "    EGL_EXTENSIONS: %s",
-                     eglQueryString(params->egl_display, EGL_EXTENSIONS));
-        }
-#endif
+    ext_arr_t exts = {0};
+    if (pl_gl->major >= 3) {
+        gl->GetIntegerv(GL_NUM_EXTENSIONS, &exts.num);
+        PL_ARRAY_RESIZE(pl_gl, exts, exts.num);
+        for (int i = 0; i < exts.num; i++)
+            exts.elem[i] = (const char *) gl->GetStringi(GL_EXTENSIONS, i);
+    } else {
+        add_exts_str(pl_gl, &exts, (const char *) gl->GetString(GL_EXTENSIONS));
     }
 
-    if (!params->allow_software && gl_is_software()) {
+    if (pl_msg_test(log, PL_LOG_DEBUG)) {
+        PL_DEBUG(p, "    GL_EXTENSIONS:");
+        for (int i = 0; i < exts.num; i++)
+            PL_DEBUG(p, "        %s", exts.elem[i]);
+    }
+
+    if (params->egl_display) {
+        pl_static_mutex_lock(&glad_egl_mutex);
+        if (params->get_proc_addr_ex) {
+            ok = gladLoadEGLUserPtr(params->egl_display, params->get_proc_addr_ex,
+                                    params->proc_ctx);
+        } else if (params->get_proc_addr) {
+            ok = gladLoadEGL(params->egl_display, params->get_proc_addr);
+        } else {
+#ifdef PL_HAVE_GL_PROC_ADDR
+            ok = gladLoaderLoadEGL(params->egl_display);
+#else
+            pl_unreachable();
+#endif
+        }
+        pl_static_mutex_unlock(&glad_egl_mutex);
+
+        if (!ok) {
+            PL_FATAL(p, "Failed loading EGL functions - double check EGLDisplay?");
+            goto error;
+        }
+
+        int start = exts.num;
+        add_exts_str(pl_gl, &exts, eglQueryString(params->egl_display,
+                                                  EGL_EXTENSIONS));
+        if (exts.num > start) {
+            PL_DEBUG(p, "    EGL_EXTENSIONS:");
+            for (int i = start; i < exts.num; i++)
+                PL_DEBUG(p, "        %s", exts.elem[i]);
+        }
+    }
+
+    pl_gl->extensions = exts.elem;
+    pl_gl->num_extensions = exts.num;
+
+    if (!params->allow_software && gl_is_software(pl_gl)) {
         PL_FATAL(p, "OpenGL context is suspected to be a software rasterizer, "
                  "but `allow_software` is false.");
         goto error;
     }
 
     if (params->debug) {
-        if (epoxy_has_gl_extension("GL_KHR_debug")) {
-            glDebugMessageCallback(debug_cb, log);
-            glEnable(GL_DEBUG_OUTPUT);
+        if (pl_opengl_has_ext(pl_gl, "GL_KHR_debug")) {
+            gl->DebugMessageCallback(debug_cb, log);
+            gl->Enable(GL_DEBUG_OUTPUT);
             p->is_debug = true;
         } else {
             PL_WARN(p, "OpenGL debugging requested, but GL_KHR_debug is not "
                     "available... ignoring!");
         }
 
-#ifdef EPOXY_HAS_EGL
-        if (params->egl_display && epoxy_has_egl_extension(params->egl_display, "EGL_KHR_debug")) {
+        if (params->egl_display && pl_opengl_has_ext(pl_gl, "EGL_KHR_debug")) {
             static const EGLAttrib attribs[] = {
                 // Enable everything under the sun, because the `pl_ctx` log
                 // level may change at runtime.
@@ -187,7 +286,6 @@ pl_opengl pl_opengl_create(pl_log log, const struct pl_opengl_params *params)
             eglLabelObjectKHR(NULL, EGL_OBJECT_THREAD_KHR, NULL, (void *) log);
             p->is_debug_egl = true;
         }
-#endif // EPOXY_HAS_EGL
     }
 
     pl_gl->gpu = pl_gpu_create_gl(log, pl_gl, params);
@@ -202,6 +300,12 @@ pl_opengl pl_opengl_create(pl_log log, const struct pl_opengl_params *params)
                 params->max_glsl_version, glsl->version);
     }
 
+    // The layout of variables inside an SSBO is implementation-defined,
+    // disable SSBO if layout qualifier is not supported.
+    struct pl_gpu_t *gpu = (struct pl_gpu_t *) pl_gl->gpu;
+    if (gpu->glsl.version < 140)
+        gpu->limits.max_ssbo_size = 0;
+
     gl_release_current(pl_gl);
     return pl_gl;
 
@@ -212,9 +316,9 @@ error:
     return NULL;
 }
 
-bool gl_make_current(pl_opengl gl)
+bool gl_make_current(pl_opengl pl_gl)
 {
-    struct priv *p = PL_PRIV(gl);
+    struct gl_ctx *p = PL_PRIV(pl_gl);
     pl_mutex_lock(&p->lock);
     if (!p->count && p->params.make_current) {
         if (!p->params.make_current(p->params.priv)) {
@@ -228,9 +332,9 @@ bool gl_make_current(pl_opengl gl)
     return true;
 }
 
-void gl_release_current(pl_opengl gl)
+void gl_release_current(pl_opengl pl_gl)
 {
-    struct priv *p = PL_PRIV(gl);
+    struct gl_ctx *p = PL_PRIV(pl_gl);
     p->count--;
     if (!p->count && p->params.release_current)
         p->params.release_current(p->params.priv);
