@@ -20,7 +20,7 @@
 #include "formats.h"
 #include "utils.h"
 
-#if defined(PL_HAVE_UNIX) && defined(EPOXY_HAS_EGL)
+#ifdef PL_HAVE_UNIX
 static bool supported_fourcc(struct pl_gl *p, EGLint fourcc)
 {
     for (int i = 0; i < p->egl_formats.num; ++i)
@@ -31,7 +31,7 @@ static bool supported_fourcc(struct pl_gl *p, EGLint fourcc)
 #endif
 
 #define FMT(_name, bits, ftype, _caps)               \
-    (struct pl_fmt) {                                \
+    (struct pl_fmt_t) {                              \
         .name = _name,                               \
         .type = PL_FMT_##ftype,                      \
         .caps = (enum pl_fmt_caps) (_caps),          \
@@ -208,10 +208,10 @@ const struct gl_format formats_basic_vertex[] = {
 
 static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
 {
-    struct pl_gpu *gpu = (struct pl_gpu *) pgpu;
+    struct pl_gpu_t *gpu = (struct pl_gpu_t *) pgpu;
     struct pl_gl *p = PL_PRIV(gpu);
 
-    struct pl_fmt *fmt = pl_alloc_obj(gpu, fmt, gl_fmt);
+    struct pl_fmt_t *fmt = pl_alloc_obj(gpu, fmt, gl_fmt);
     const struct gl_format **fmtp = PL_PRIV(fmt);
     *fmt = gl_fmt->tmpl;
     *fmtp = gl_fmt;
@@ -294,7 +294,7 @@ static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
     fmt->fourcc = pl_fmt_fourcc(fmt);
     pl_assert(fmt->glsl_type);
 
-#if defined(PL_HAVE_UNIX) && defined(EPOXY_HAS_EGL)
+#ifdef PL_HAVE_UNIX
     if (p->has_modifiers && fmt->fourcc && supported_fourcc(p, fmt->fourcc)) {
         int num_mods = 0;
         bool ok = eglQueryDmaBufModifiersEXT(p->egl_dpy, fmt->fourcc,
@@ -377,11 +377,11 @@ static void add_format(pl_gpu pgpu, const struct gl_format *gl_fmt)
             add_format(gpu, &formats[i]);                   \
     } while (0)
 
-bool gl_setup_formats(struct pl_gpu *gpu)
+bool gl_setup_formats(struct pl_gpu_t *gpu)
 {
     struct pl_gl *p = PL_PRIV(gpu);
 
-#if defined(PL_HAVE_UNIX) && defined(EPOXY_HAS_EGL)
+#ifdef PL_HAVE_UNIX
     if (p->has_modifiers) {
         EGLint num_formats = 0;
         bool ok = eglQueryDmaBufFormatsEXT(p->egl_dpy, 0, NULL,
@@ -416,9 +416,9 @@ bool gl_setup_formats(struct pl_gpu *gpu)
     if (p->gl_ver >= 21) {
         // If we have a reasonable set of extensions, we can enable most
         // things. Otherwise, pick simple fallback formats
-        if (epoxy_has_gl_extension("GL_ARB_texture_float") &&
-            epoxy_has_gl_extension("GL_ARB_texture_rg") &&
-            epoxy_has_gl_extension("GL_ARB_framebuffer_object"))
+        if (pl_opengl_has_ext(p->gl, "GL_ARB_texture_float") &&
+            pl_opengl_has_ext(p->gl, "GL_ARB_texture_rg") &&
+            pl_opengl_has_ext(p->gl, "GL_ARB_framebuffer_object"))
         {
             DO_FORMATS(formats_norm8);
             DO_FORMATS(formats_bgra8);
@@ -437,17 +437,17 @@ bool gl_setup_formats(struct pl_gpu *gpu)
         // GLES 3.0 has some basic formats, with framebuffers for float16
         // depending on GL_EXT_color_buffer_(half_)float support
         DO_FORMATS(formats_norm8);
-        if (epoxy_has_gl_extension("GL_EXT_texture_norm16")) {
+        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_norm16")) {
             DO_FORMATS(formats_norm16);
             DO_FORMATS(formats_rgb16_fallback);
         }
-        if (epoxy_has_gl_extension("GL_EXT_texture_format_BGRA8888"))
+        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_format_BGRA8888"))
             DO_FORMATS(formats_bgra_gles);
-        if (epoxy_has_gl_extension("GL_EXT_texture_integer"))
+        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_integer"))
             DO_FORMATS(formats_uint);
         DO_FORMATS(formats_basic_vertex);
-        if (p->gles_ver >= 32 || epoxy_has_gl_extension("GL_EXT_color_buffer_half_float")
-                              || epoxy_has_gl_extension("GL_EXT_color_buffer_float")) {
+        if (p->gles_ver >= 32 || pl_opengl_has_ext(p->gl, "GL_EXT_color_buffer_half_float")
+                              || pl_opengl_has_ext(p->gl, "GL_EXT_color_buffer_float")) {
             DO_FORMATS(formats_float16_fbo);
         } else {
             DO_FORMATS(formats_float16_fallback);
@@ -460,10 +460,10 @@ bool gl_setup_formats(struct pl_gpu *gpu)
         // float16 depending on GL_EXT_texture_norm16 being present
         DO_FORMATS(formats_legacy_gles2);
         DO_FORMATS(formats_basic_vertex);
-        if (epoxy_has_gl_extension("GL_EXT_texture_rg")) {
+        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_rg")) {
             DO_FORMATS(formats_norm8);
         }
-        if (epoxy_has_gl_extension("GL_EXT_texture_format_BGRA8888")) {
+        if (pl_opengl_has_ext(p->gl, "GL_EXT_texture_format_BGRA8888")) {
             DO_FORMATS(formats_bgra_gles);
         }
         goto done;
