@@ -18,18 +18,22 @@
 #pragma once
 
 #include <windows.h>
+#include <stdint.h>
 #include <errno.h>
 
+#include <pl_assert.h>
+
 typedef CRITICAL_SECTION   pl_mutex;
 typedef CONDITION_VARIABLE pl_cond;
 
 static inline int pl_mutex_init_type_internal(pl_mutex *mutex, enum pl_mutex_type mtype)
 {
+    (void) mtype;
     return !InitializeCriticalSectionEx(mutex, 0, 0);
 }
 
 #define pl_mutex_init_type(mutex, mtype) \
-    PL_CHECK_ERR(pl_mutex_init_type_internal(mutex, mtype))
+    pl_assert(!pl_mutex_init_type_internal(mutex, mtype))
 
 static inline int pl_mutex_destroy(pl_mutex *mutex)
 {
@@ -58,6 +62,7 @@ static inline int pl_cond_init(pl_cond *cond)
 static inline int pl_cond_destroy(pl_cond *cond)
 {
     // condition variables are not destroyed
+    (void) cond;
     return 0;
 }
 
@@ -83,7 +88,11 @@ static inline int pl_cond_timedwait(pl_cond *cond, pl_mutex *mutex, uint64_t tim
     if (timeout == UINT64_MAX)
         return pl_cond_wait(cond, mutex);
 
-    BOOL bRet = SleepConditionVariableCS(cond, mutex, PL_MIN(timeout / 1000000LLU, INFINITE - 1));
+    timeout /= UINT64_C(1000000);
+    if (timeout > INFINITE - 1)
+        timeout = INFINITE - 1;
+
+    BOOL bRet = SleepConditionVariableCS(cond, mutex, timeout);
     if (bRet == FALSE)
     {
         if (GetLastError() == ERROR_TIMEOUT)
