@@ -25,11 +25,14 @@
 pl_shader pl_shader_alloc(pl_log log, const struct pl_shader_params *params)
 {
     pl_shader sh = pl_alloc_ptr(NULL, sh);
-    *sh = (struct pl_shader) {
+    *sh = (struct pl_shader_t) {
         .log = log,
         .mutable = true,
     };
 
+    for (int i = 0; i < PL_ARRAY_SIZE(sh->buffers); i++)
+        sh->buffers[i] = pl_str_builder_alloc(sh);
+
     // Ensure there's always at least one `tmp` object
     PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
 
@@ -39,24 +42,31 @@ pl_shader pl_shader_alloc(pl_log log, const struct pl_shader_params *params)
     return sh;
 }
 
+static void sh_obj_deref(pl_shader_obj obj);
+
+static void sh_deref(pl_shader sh)
+{
+    for (int i = 0; i < sh->tmp.num; i++)
+        pl_ref_deref(&sh->tmp.elem[i]);
+    for (int i = 0; i < sh->obj.num; i++)
+        sh_obj_deref(sh->obj.elem[i]);
+}
+
 void pl_shader_free(pl_shader *psh)
 {
     pl_shader sh = *psh;
     if (!sh)
         return;
 
-    for (int i = 0; i < sh->tmp.num; i++)
-        pl_ref_deref(&sh->tmp.elem[i]);
-
+    sh_deref(sh);
     pl_free_ptr(psh);
 }
 
 void pl_shader_reset(pl_shader sh, const struct pl_shader_params *params)
 {
-    for (int i = 0; i < sh->tmp.num; i++)
-        pl_ref_deref(&sh->tmp.elem[i]);
+    sh_deref(sh);
 
-    struct pl_shader new = {
+    struct pl_shader_t new = {
         .log = sh->log,
         .mutable = true,
 
@@ -73,8 +83,9 @@ void pl_shader_reset(pl_shader sh, const struct pl_shader_params *params)
         new.res.params = *params;
 
     // Preserve buffer allocations
+    memcpy(new.buffers, sh->buffers, sizeof(new.buffers));
     for (int i = 0; i < PL_ARRAY_SIZE(new.buffers); i++)
-        new.buffers[i] = (pl_str) { .buf = sh->buffers[i].buf };
+        pl_str_builder_reset(new.buffers[i]);
 
     *sh = new;
     PL_ARRAY_APPEND(sh, sh->tmp, pl_ref_new(NULL));
@@ -198,6 +209,33 @@ ident_t sh_var(pl_shader sh, struct pl_shader_var sv)
     return (ident_t) sv.var.name;
 }
 
+ident_t sh_var_int(pl_shader sh, const char *name, int val, bool dynamic)
+{
+    return sh_var(sh, (struct pl_shader_var) {
+        .var     = pl_var_int(name),
+        .data    = &val,
+        .dynamic = dynamic,
+    });
+}
+
+ident_t sh_var_uint(pl_shader sh, const char *name, unsigned int val, bool dynamic)
+{
+    return sh_var(sh, (struct pl_shader_var) {
+        .var     = pl_var_uint(name),
+        .data    = &val,
+        .dynamic = dynamic,
+    });
+}
+
+ident_t sh_var_float(pl_shader sh, const char *name, float val, bool dynamic)
+{
+    return sh_var(sh, (struct pl_shader_var) {
+        .var     = pl_var_float(name),
+        .data    = &val,
+        .dynamic = dynamic,
+    });
+}
+
 static void merge_access(enum pl_desc_access *a, enum pl_desc_access b)
 {
     if (*a != b)
@@ -259,9 +297,11 @@ ident_t sh_const(pl_shader sh, struct pl_shader_const sc)
 
     pl_gpu gpu = SH_GPU(sh);
     if (gpu && gpu->limits.max_constants) {
-        sc.data = pl_memdup(SH_TMP(sh), sc.data, pl_var_type_size(sc.type));
-        PL_ARRAY_APPEND(sh, sh->consts, sc);
-        return (ident_t) sc.name;
+        if (!sc.compile_time || gpu->limits.array_size_constants) {
+            sc.data = pl_memdup(SH_TMP(sh), sc.data, pl_var_type_size(sc.type));
+            PL_ARRAY_APPEND(sh, sh->consts, sc);
+            return (ident_t) sc.name;
+        }
     }
 
     // Fallback for GPUs without specialization constants
@@ -310,7 +350,6 @@ ident_t sh_const_float(pl_shader sh, const char *name, float val)
     });
 }
 
-
 ident_t sh_attr_vec2(pl_shader sh, const char *name,
                      const struct pl_rect2df *rc)
 {
@@ -457,22 +496,6 @@ size_t sh_buf_desc_size(const struct pl_shader_desc *buf_desc)
     return last->layout.offset + last->layout.size;
 }
 
-void sh_append(pl_shader sh, enum pl_shader_buf buf, const char *fmt, ...)
-{
-    pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
-
-    va_list ap;
-    va_start(ap, fmt);
-    pl_str_append_vasprintf_c(sh, &sh->buffers[buf], fmt, ap);
-    va_end(ap);
-}
-
-void sh_append_str(pl_shader sh, enum pl_shader_buf buf, pl_str str)
-{
-    pl_assert(buf >= 0 && buf < SH_BUF_COUNT);
-    pl_str_append(sh, &sh->buffers[buf], str);
-}
-
 void sh_describef(pl_shader sh, const char *fmt, ...)
 {
     va_list ap;
@@ -540,8 +563,8 @@ ident_t sh_subpass(pl_shader sh, const pl_shader sub)
     sh->output_h = res_h;
 
     // Append the prelude and header
-    pl_str_append(sh, &sh->buffers[SH_BUF_PRELUDE], sub->buffers[SH_BUF_PRELUDE]);
-    pl_str_append(sh, &sh->buffers[SH_BUF_HEADER],  sub->buffers[SH_BUF_HEADER]);
+    pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sub->buffers[SH_BUF_PRELUDE]);
+    pl_str_builder_concat(sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_HEADER]);
 
     // Append the body as a new header function
     ident_t name = sh_fresh(sh, "sub");
@@ -553,7 +576,7 @@ ident_t sh_subpass(pl_shader sh, const pl_shader sub)
     } else {
         GLSLH("%s %s(%s) {\n", outsigs[sub->res.output], name, insigs[sub->res.input]);
     }
-    pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_BODY]);
+    pl_str_builder_concat(sh->buffers[SH_BUF_HEADER], sub->buffers[SH_BUF_BODY]);
     GLSLH("%s\n}\n\n", retvals[sub->res.output]);
 
     // Copy over all of the descriptors etc.
@@ -568,62 +591,38 @@ ident_t sh_subpass(pl_shader sh, const pl_shader sub)
     return name;
 }
 
-// Finish the current shader body and return its function name
-static ident_t sh_split(pl_shader sh)
+pl_str_builder sh_finalize_internal(pl_shader sh)
 {
     pl_assert(sh->mutable);
-
-    // Concatenate the body onto the head as a new function
-    ident_t name = sh_fresh(sh, "main");
-    if (sh->res.input == PL_SHADER_SIG_SAMPLER) {
-        pl_assert(sh->sampler_prefix);
-        GLSLH("%s %s(%c%s src_tex, vec2 tex_coord) {\n",
-              outsigs[sh->res.output], name,
-              sh->sampler_prefix, samplers2D[sh->sampler_type]);
-    } else {
-        GLSLH("%s %s(%s) {\n", outsigs[sh->res.output], name, insigs[sh->res.input]);
-    }
-
-    if (sh->buffers[SH_BUF_BODY].len) {
-        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_BODY]);
-        sh->buffers[SH_BUF_BODY].len = 0;
-        sh->buffers[SH_BUF_BODY].buf[0] = '\0'; // for sanity / efficiency
-    }
-
-    if (sh->buffers[SH_BUF_FOOTER].len) {
-        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], sh->buffers[SH_BUF_FOOTER]);
-        sh->buffers[SH_BUF_FOOTER].len = 0;
-        sh->buffers[SH_BUF_FOOTER].buf[0] = '\0';
-    }
-
-    GLSLH("%s\n}\n\n", retvals[sh->res.output]);
-    return name;
-}
-
-const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
-{
     if (sh->failed)
         return NULL;
 
-    if (!sh->mutable)
-        return &sh->res;
-
-    // Split the shader. This finalizes the body and adds it to the header
-    sh->res.name = sh_split(sh);
-
     // Padding for readability
     GLSLP("\n");
 
-    // Concatenate the header onto the prelude to form the final output
-    pl_str *glsl = &sh->buffers[SH_BUF_PRELUDE];
-    pl_str_append(sh, glsl, sh->buffers[SH_BUF_HEADER]);
+    // Concatenate everything onto the prelude to form the final output
+    pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_HEADER]);
+
+    sh->res.name = sh_fresh(sh, "main");
+    if (sh->res.input == PL_SHADER_SIG_SAMPLER) {
+        pl_assert(sh->sampler_prefix);
+        GLSLP("%s %s(%c%s src_tex, vec2 tex_coord) {\n",
+              outsigs[sh->res.output], sh->res.name,
+              sh->sampler_prefix, samplers2D[sh->sampler_type]);
+    } else {
+        GLSLP("%s %s(%s) {\n", outsigs[sh->res.output], sh->res.name, insigs[sh->res.input]);
+    }
+
+    pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_BODY]);
+    pl_str_builder_concat(sh->buffers[SH_BUF_PRELUDE], sh->buffers[SH_BUF_FOOTER]);
+    GLSLP("%s\n}\n\n", retvals[sh->res.output]);
 
     // Generate the pretty description
     sh->res.description = "(unknown shader)";
     if (sh->steps.num) {
-        // Reuse this buffer
-        pl_str *desc = &sh->buffers[SH_BUF_BODY];
-        desc->len = 0;
+        // Reuse this builder
+        pl_str_builder desc = sh->buffers[SH_BUF_BODY];
+        pl_str_builder_reset(desc);
 
         for (int i = 0; i < sh->steps.num; i++) {
             const char *step = sh->steps.elem[i];
@@ -641,13 +640,13 @@ const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
             }
 
             if (i > 0)
-                pl_str_append(sh, desc, pl_str0(", "));
-            pl_str_append(sh, desc, pl_str0(step));
+                pl_str_builder_const_str(desc, ", ");
+            pl_str_builder_str0(desc, step);
             if (count > 1)
-                pl_str_append_asprintf(sh, desc, " x%d", count);
+                pl_str_builder_printf_c(desc, " x%d", count);
         }
 
-        sh->res.description = (char *) desc->buf;
+        sh->res.description = (char *) pl_str_builder_exec(desc).buf;
     }
 
     // Set the vas/vars/descs
@@ -661,10 +660,23 @@ const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
     sh->res.num_constants = sh->consts.num;
     sh->res.steps = sh->steps.elem;
     sh->res.num_steps = sh->steps.num;
-
-    // Update the result pointer and return
-    sh->res.glsl = (char *) glsl->buf;
     sh->mutable = false;
+    return sh->buffers[SH_BUF_PRELUDE];
+}
+
+const struct pl_shader_res *pl_shader_finalize(pl_shader sh)
+{
+    pl_str_builder glsl = NULL;
+    if (sh->mutable) {
+        glsl = sh_finalize_internal(sh);
+        if (!glsl)
+            return NULL;
+    }
+
+    pl_assert(!sh->mutable);
+    if (!sh->res.glsl)
+        sh->res.glsl = (char *) pl_str_builder_exec(glsl).buf;
+
     return &sh->res;
 }
 
@@ -713,17 +725,25 @@ bool sh_require(pl_shader sh, enum pl_shader_sig insig, int w, int h)
     return true;
 }
 
+static void sh_obj_deref(pl_shader_obj obj)
+{
+    if (!pl_rc_deref(&obj->rc))
+        return;
+
+    if (obj->uninit)
+        obj->uninit(obj->gpu, obj->priv);
+
+    pl_free(obj);
+}
+
 void pl_shader_obj_destroy(pl_shader_obj *ptr)
 {
     pl_shader_obj obj = *ptr;
     if (!obj)
         return;
 
-    if (obj->uninit)
-        obj->uninit(obj->gpu, obj->priv);
-
+    sh_obj_deref(obj);
     *ptr = NULL;
-    pl_free(obj);
 }
 
 void *sh_require_obj(pl_shader sh, pl_shader_obj *ptr,
@@ -747,12 +767,16 @@ void *sh_require_obj(pl_shader sh, pl_shader_obj *ptr,
 
     if (!obj) {
         obj = pl_zalloc_ptr(NULL, obj);
+        pl_rc_init(&obj->rc);
         obj->gpu = SH_GPU(sh);
         obj->type = type;
         obj->priv = pl_zalloc(obj, priv_size);
         obj->uninit = uninit;
     }
 
+    PL_ARRAY_APPEND(sh, sh->obj, obj);
+    pl_rc_ref(&obj->rc);
+
     *ptr = obj;
     return obj->priv;
 }
@@ -836,425 +860,6 @@ ident_t sh_prng(pl_shader sh, bool temporal, ident_t *p_state)
     return res;
 }
 
-// Defines a LUT position helper macro. This translates from an absolute texel
-// scale (0.0 - 1.0) to the texture coordinate scale for the corresponding
-// sample in a texture of dimension `lut_size`.
-static ident_t sh_lut_pos(pl_shader sh, int lut_size)
-{
-    ident_t name = sh_fresh(sh, "LUT_POS");
-    GLSLH("#define %s(x) mix(%s, %s, (x)) \n",
-          name, SH_FLOAT(0.5 / lut_size), SH_FLOAT(1.0 - 0.5 / lut_size));
-    return name;
-}
-
-struct sh_lut_obj {
-    enum sh_lut_method method;
-    enum pl_var_type type;
-    pl_fmt fmt;
-    bool linear;
-    int width, height, depth, comps;
-    uint64_t signature;
-    bool error; // reset if params change
-
-    // weights, depending on the method
-    pl_tex tex;
-    pl_str str;
-    void *data;
-};
-
-static void sh_lut_uninit(pl_gpu gpu, void *ptr)
-{
-    struct sh_lut_obj *lut = ptr;
-    pl_tex_destroy(gpu, &lut->tex);
-    pl_free(lut->str.buf);
-    pl_free(lut->data);
-
-    *lut = (struct sh_lut_obj) {0};
-}
-
-// Maximum number of floats to embed as a literal array (when using SH_LUT_AUTO)
-#define SH_LUT_MAX_LITERAL_SOFT 64
-#define SH_LUT_MAX_LITERAL_HARD 256
-
-ident_t sh_lut(pl_shader sh, const struct sh_lut_params *params)
-{
-    pl_gpu gpu = SH_GPU(sh);
-    void *tmp = NULL;
-
-    pl_assert(params->width > 0 && params->height >= 0 && params->depth >= 0);
-    pl_assert(params->comps > 0);
-    pl_assert(params->type);
-    pl_assert(!params->linear || params->type == PL_VAR_FLOAT);
-
-    int sizes[] = { params->width, params->height, params->depth };
-    int size = params->width * PL_DEF(params->height, 1) * PL_DEF(params->depth, 1);
-    int dims = params->depth ? 3 : params->height ? 2 : 1;
-
-    int texdim = 0;
-    uint32_t max_tex_dim[] = {
-        gpu ? gpu->limits.max_tex_1d_dim : 0,
-        gpu ? gpu->limits.max_tex_2d_dim : 0,
-        (gpu && gpu->glsl.version > 100) ? gpu->limits.max_tex_3d_dim : 0,
-    };
-
-    struct sh_lut_obj *lut = SH_OBJ(sh, params->object, PL_SHADER_OBJ_LUT,
-                                    struct sh_lut_obj, sh_lut_uninit);
-
-    if (!lut)
-        return NULL;
-
-    bool update = params->update || lut->signature != params->signature ||
-                  params->type != lut->type || params->fmt != lut->fmt ||
-                  params->linear != lut->linear ||
-                  params->width != lut->width || params->height != lut->height ||
-                  params->depth != lut->depth || params->comps != lut->comps;
-
-    if (lut->error && !update)
-        return NULL; // suppress error spam until something changes
-
-    // Try picking the right number of dimensions for the texture LUT. This
-    // allows e.g. falling back to 2D textures if 1D textures are unsupported.
-    for (int d = dims; d <= PL_ARRAY_SIZE(max_tex_dim); d++) {
-        // For a given dimension to be compatible, all coordinates need to be
-        // within the maximum texture size for that dimension
-        for (int i = 0; i < d; i++) {
-            if (sizes[i] > max_tex_dim[d - 1])
-                goto next_dim;
-        }
-
-        // All dimensions are compatible, so pick this texture dimension
-        texdim = d;
-        break;
-
-next_dim: ; // `continue` out of the inner loop
-    }
-
-    static const enum pl_fmt_type fmt_type[PL_VAR_TYPE_COUNT] = {
-        [PL_VAR_SINT]   = PL_FMT_SINT,
-        [PL_VAR_UINT]   = PL_FMT_UINT,
-        [PL_VAR_FLOAT]  = PL_FMT_FLOAT,
-    };
-
-    enum pl_fmt_caps texcaps = PL_FMT_CAP_SAMPLEABLE;
-    if (params->linear)
-        texcaps |= PL_FMT_CAP_LINEAR;
-
-    pl_fmt texfmt = params->fmt;
-    if (texfmt) {
-        bool ok;
-        switch (texfmt->type) {
-        case PL_FMT_SINT: ok = params->type == PL_VAR_SINT; break;
-        case PL_FMT_UINT: ok = params->type == PL_VAR_UINT; break;
-        default:          ok = params->type == PL_VAR_FLOAT; break;
-        }
-
-        if (!ok) {
-            PL_ERR(sh, "Specified texture format '%s' does not match LUT "
-                   "data type!", texfmt->name);
-            goto error;
-        }
-
-        if (~texfmt->caps & texcaps) {
-            PL_ERR(sh, "Specified texture format '%s' does not match "
-                   "required capabilities 0x%x!\n", texfmt->name, texcaps);
-            goto error;
-        }
-    }
-
-    if (texdim && !texfmt) {
-        texfmt = pl_find_fmt(gpu, fmt_type[params->type], params->comps,
-                             params->type == PL_VAR_FLOAT ? 16 : 32,
-                             pl_var_type_size(params->type) * 8,
-                             texcaps);
-    }
-
-    enum sh_lut_method method = params->method;
-
-    // The linear sampling code currently only supports 1D linear interpolation
-    if (params->linear && dims > 1) {
-        if (texfmt) {
-            method = SH_LUT_TEXTURE;
-        } else {
-            PL_ERR(sh, "Can't emulate linear LUTs for 2D/3D LUTs and no "
-                  "texture support available!");
-            goto error;
-        }
-    }
-
-    bool can_uniform = gpu && gpu->limits.max_variable_comps >= size * params->comps;
-    bool can_literal = sh_glsl(sh).version > 110; // needed for literal arrays
-    can_literal &= size <= SH_LUT_MAX_LITERAL_HARD && !params->dynamic;
-
-    // Deselect unsupported methods
-    if (method == SH_LUT_UNIFORM && !can_uniform)
-        method = SH_LUT_AUTO;
-    if (method == SH_LUT_LITERAL && !can_literal)
-        method = SH_LUT_AUTO;
-    if (method == SH_LUT_TEXTURE && !texfmt)
-        method = SH_LUT_AUTO;
-
-    // Sorted by priority
-    if (!method && can_literal && size <= SH_LUT_MAX_LITERAL_SOFT)
-        method = SH_LUT_LITERAL;
-    if (!method && texfmt)
-        method = SH_LUT_TEXTURE;
-    if (!method && can_uniform)
-        method = SH_LUT_UNIFORM;
-    if (!method && can_literal)
-        method = SH_LUT_LITERAL;
-
-    if (!method) {
-        PL_ERR(sh, "Can't generate LUT: no compatible methods!");
-        goto error;
-    }
-
-    // Reinitialize the existing LUT if needed
-    update |= method != lut->method;
-    if (update) {
-        PL_DEBUG(sh, "LUT cache invalidated, regenerating..");
-        size_t buf_size = size * params->comps * pl_var_type_size(params->type);
-        tmp = pl_zalloc(NULL, buf_size);
-        params->fill(tmp, params);
-
-        switch (method) {
-        case SH_LUT_TEXTURE: {
-            if (!texdim) {
-                PL_ERR(sh, "Texture LUT exceeds texture dimensions!");
-                goto error;
-            }
-
-            if (!texfmt) {
-                PL_ERR(sh, "Found no compatible texture format for LUT!");
-                goto error;
-            }
-
-            struct pl_tex_params tex_params = {
-                .w              = params->width,
-                .h              = PL_DEF(params->height, texdim >= 2 ? 1 : 0),
-                .d              = PL_DEF(params->depth,  texdim >= 3 ? 1 : 0),
-                .format         = texfmt,
-                .sampleable     = true,
-                .host_writable  = params->dynamic,
-                .initial_data   = params->dynamic ? NULL : tmp,
-                .debug_tag      = PL_DEBUG_TAG,
-            };
-
-            bool ok;
-            if (params->dynamic) {
-                ok = pl_tex_recreate(gpu, &lut->tex, &tex_params);
-                if (ok) {
-                    ok = pl_tex_upload(gpu, pl_tex_transfer_params(
-                        .tex = lut->tex,
-                        .ptr = tmp,
-                    ));
-                }
-            } else {
-                // Can't use pl_tex_recreate because of `initial_data`
-                pl_tex_destroy(gpu, &lut->tex);
-                lut->tex = pl_tex_create(gpu, &tex_params);
-                ok = lut->tex;
-            }
-
-            if (!ok) {
-                PL_ERR(sh, "Failed creating LUT texture!");
-                goto error;
-            }
-            break;
-        }
-
-        case SH_LUT_UNIFORM:
-            pl_free(lut->data);
-            lut->data = tmp; // re-use `tmp`
-            tmp = NULL;
-            break;
-
-        case SH_LUT_LITERAL: {
-            lut->str.len = 0;
-            static const char prefix[PL_VAR_TYPE_COUNT] = {
-                [PL_VAR_SINT]   = 'i',
-                [PL_VAR_UINT]   = 'u',
-                [PL_VAR_FLOAT]  = ' ',
-            };
-
-            for (int i = 0; i < size * params->comps; i += params->comps) {
-                if (i > 0)
-                    pl_str_append_asprintf_c(lut, &lut->str, ",");
-                if (params->comps > 1) {
-                    pl_str_append_asprintf_c(lut, &lut->str, "%cvec%d(",
-                                             prefix[params->type], params->comps);
-                }
-                for (int c = 0; c < params->comps; c++) {
-                    switch (params->type) {
-                    case PL_VAR_FLOAT:
-                        pl_str_append_asprintf_c(lut, &lut->str, "%s%f",
-                                                 c > 0 ? "," : "",
-                                                 ((float *) tmp)[i+c]);
-                        break;
-                    case PL_VAR_UINT:
-                        pl_str_append_asprintf_c(lut, &lut->str, "%s%u",
-                                                 c > 0 ? "," : "",
-                                                 ((unsigned int *) tmp)[i+c]);
-                        break;
-                    case PL_VAR_SINT:
-                        pl_str_append_asprintf_c(lut, &lut->str, "%s%d",
-                                                 c > 0 ? "," : "",
-                                                 ((int *) tmp)[i+c]);
-                        break;
-                    case PL_VAR_INVALID:
-                    case PL_VAR_TYPE_COUNT:
-                        pl_unreachable();
-                    }
-                }
-                if (params->comps > 1)
-                    pl_str_append_asprintf_c(lut, &lut->str, ")");
-            }
-            break;
-        }
-
-        case SH_LUT_AUTO:
-            pl_unreachable();
-        }
-
-        lut->method = method;
-        lut->type = params->type;
-        lut->fmt = params->fmt;
-        lut->linear = params->linear;
-        lut->width = params->width;
-        lut->height = params->height;
-        lut->depth = params->depth;
-        lut->comps = params->comps;
-        lut->signature = params->signature;
-    }
-
-    // Done updating, generate the GLSL
-    ident_t name = sh_fresh(sh, "lut");
-    ident_t arr_name = NULL;
-
-    static const char * const swizzles[] = {"x", "xy", "xyz", "xyzw"};
-    static const char * const vartypes[PL_VAR_TYPE_COUNT][4] = {
-        [PL_VAR_SINT] = { "int", "ivec2", "ivec3", "ivec4" },
-        [PL_VAR_UINT] = { "uint", "uvec2", "uvec3", "uvec4" },
-        [PL_VAR_FLOAT] = { "float", "vec2", "vec3", "vec4" },
-    };
-
-    switch (method) {
-    case SH_LUT_TEXTURE: {
-        assert(texdim);
-        ident_t tex = sh_desc(sh, (struct pl_shader_desc) {
-            .desc = {
-                .name = "weights",
-                .type = PL_DESC_SAMPLED_TEX,
-            },
-            .binding = {
-                .object = lut->tex,
-                .sample_mode = params->linear ? PL_TEX_SAMPLE_LINEAR
-                                              : PL_TEX_SAMPLE_NEAREST,
-            }
-        });
-
-        // texelFetch requires GLSL >= 130, so fall back to the linear code
-        if (params->linear || gpu->glsl.version < 130) {
-            ident_t pos_macros[PL_ARRAY_SIZE(sizes)] = {0};
-            for (int i = 0; i < dims; i++)
-                pos_macros[i] = sh_lut_pos(sh, sizes[i]);
-
-            GLSLH("#define %s(pos) (%s(%s, %s(\\\n",
-                  name, sh_tex_fn(sh, lut->tex->params),
-                  tex, vartypes[PL_VAR_FLOAT][texdim - 1]);
-
-            for (int i = 0; i < texdim; i++) {
-                char sep = i == 0 ? ' ' : ',';
-                if (pos_macros[i]) {
-                    if (dims > 1) {
-                        GLSLH("   %c%s(%s(pos).%c)\\\n", sep, pos_macros[i],
-                              vartypes[PL_VAR_FLOAT][dims - 1], "xyzw"[i]);
-                    } else {
-                        GLSLH("   %c%s(float(pos))\\\n", sep, pos_macros[i]);
-                    }
-                } else {
-                    GLSLH("   %c%f\\\n", sep, 0.5);
-                }
-            }
-            GLSLH("  )).%s)\n", swizzles[params->comps - 1]);
-        } else {
-            GLSLH("#define %s(pos) (texelFetch(%s, %s(pos",
-                  name, tex, vartypes[PL_VAR_SINT][texdim - 1]);
-
-            // Fill up extra components of the index
-            for (int i = dims; i < texdim; i++)
-                GLSLH(", 0");
-
-            GLSLH("), 0).%s)\n", swizzles[params->comps - 1]);
-        }
-
-        break;
-    }
-
-    case SH_LUT_UNIFORM:
-        arr_name = sh_var(sh, (struct pl_shader_var) {
-            .var = {
-                .name = "weights",
-                .type = params->type,
-                .dim_v = params->comps,
-                .dim_m = 1,
-                .dim_a = size,
-            },
-            .data = lut->data,
-        });
-        break;
-
-    case SH_LUT_LITERAL:
-        arr_name = sh_fresh(sh, "weights");
-        GLSLH("const %s %s[%d] = %s[](\n  ",
-              vartypes[params->type][params->comps - 1], arr_name, size,
-              vartypes[params->type][params->comps - 1]);
-        pl_str_append(sh, &sh->buffers[SH_BUF_HEADER], lut->str);
-        GLSLH(");\n");
-        break;
-
-    case SH_LUT_AUTO:
-        pl_unreachable();
-    }
-
-    if (arr_name) {
-        GLSLH("#define %s(pos) (%s[int((pos)%s)\\\n",
-              name, arr_name, dims > 1 ? "[0]" : "");
-        int shift = params->width;
-        for (int i = 1; i < dims; i++) {
-            GLSLH("    + %d * int((pos)[%d])\\\n", shift, i);
-            shift *= sizes[i];
-        }
-        GLSLH("  ])\n");
-
-        if (params->linear) {
-            pl_assert(dims == 1);
-            pl_assert(params->type == PL_VAR_FLOAT);
-            ident_t arr_lut = name;
-            name = sh_fresh(sh, "lut_lin");
-            GLSLH("%s %s(float fpos) {                              \n"
-                  "    fpos = clamp(fpos, 0.0, 1.0) * %d.0;         \n"
-                  "    float fbase = floor(fpos);                   \n"
-                  "    float fceil = ceil(fpos);                    \n"
-                  "    float fcoord = fpos - fbase;                 \n"
-                  "    return mix(%s(fbase), %s(fceil), fcoord);    \n"
-                  "}                                                \n",
-                  vartypes[PL_VAR_FLOAT][params->comps - 1], name,
-                  size - 1,
-                  arr_lut, arr_lut);
-        }
-    }
-
-    lut->error = false;
-    pl_free(tmp);
-    pl_assert(name);
-    return name;
-
-error:
-    lut->error = true;
-    pl_free(tmp);
-    return NULL;
-}
-
 const char *sh_bvec(const pl_shader sh, int dims)
 {
     static const char *bvecs[] = {
