@@ -19,6 +19,10 @@
 
 #include "common.h"
 
+#include <libplacebo/log.h>
+#include <libplacebo/colorspace.h>
+#include <libplacebo/shaders/film_grain.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -57,22 +61,87 @@ static inline pl_log pl_test_logger(void)
     ));
 }
 
-static inline void require(bool b, const char *msg, const char *file, int line)
+#define RANDOM (rand() / (float) RAND_MAX)
+#define RANDOM_U8 ((uint8_t) (256.0 * rand() / (RAND_MAX + 1.0)))
+#define SKIP 77
+
+// Helpers for performing various checks
+#define REQUIRE(cond) do                                                        \
+{                                                                               \
+    if (!(cond)) {                                                              \
+        fprintf(stderr, "=== FAILED: '"#cond"' at "__FILE__":%d\n\n", __LINE__);\
+        exit(1);                                                                \
+    }                                                                           \
+} while (0)
+
+#define REQUIRE_CMP(a, op, b, fmt) do                                           \
+{                                                                               \
+    __typeof__(a) _va = (a), _vb = (b);                                         \
+                                                                                \
+    if (!(_va op _vb)) {                                                        \
+        fprintf(stderr, "=== FAILED: '"#a" "#op" "#b"' at "__FILE__":%d\n"      \
+                        " %-31s = %"fmt"\n"                                     \
+                        " %-31s = %"fmt"\n\n",                                  \
+                __LINE__, #a, _va, #b, _vb);                                    \
+        exit(1);                                                                \
+    }                                                                           \
+} while (0)
+
+#define REQUIRE_FEQ(a, b, epsilon) do                                           \
+{                                                                               \
+    float _va = (a);                                                            \
+    float _vb = (b);                                                            \
+    float _delta = (epsilon) * fmax(1.0, fabs(_va));                            \
+                                                                                \
+    if (fabs(_va - _vb) > _delta) {                                             \
+        fprintf(stderr, "=== FAILED: '"#a" ≈ "#b"' at "__FILE__":%d\n"          \
+                        " %-31s = %f\n"                                         \
+                        " %-31s = %f\n"                                         \
+                        " %-31s = %f\n\n",                                      \
+                __LINE__, #a, _va, #b, _vb,                                     \
+                "epsilon "#epsilon" -> max delta", _delta);                     \
+        exit(1);                                                                \
+    }                                                                           \
+} while (0)
+
+static inline void log_array(const uint8_t *a, const uint8_t *ref, size_t size)
 {
-    if (!b) {
-        fprintf(stderr, "=== FAILED: '%s' at %s:%d\n\n", msg, file, line);
+    for (size_t n = 0; n < size; n++) {
+        const char *prefix = "", *suffix = "";
+        char terminator = ' ';
+        if (a[n] != ref[n]) {
+            prefix = "\033[31;1m";
+            suffix = "\033[0m";
+        }
+        if (n+1 == size || n % 16 == 15)
+            terminator = '\n';
+        fprintf(stderr, "%s%02"PRIx8"%s%c", prefix, a[n], suffix, terminator);
+    }
+}
+
+static inline void require_memeq(const void *aptr, const void *bptr, size_t size,
+                                 const char *astr, const char *bstr,
+                                 const char *sizestr, const char *file, int line)
+{
+    const uint8_t *a = aptr, *b = bptr;
+    for (size_t i = 0; i < size; i++) {
+        if (a[i] == b[i])
+            continue;
+
+        fprintf(stderr, "=== FAILED: memcmp(%s, %s, %s) == 0 at %s:%d\n"
+                        "at position %zu: 0x%02"PRIx8" != 0x%02"PRIx8"\n\n",
+                astr, bstr, sizestr, file, line, i, a[i], b[i]);
+
+        const size_t logsize = PL_MIN(size, PL_MAX(i+2, 512));
+        fprintf(stderr, "first %zu bytes of '%s':\n", logsize, astr);
+        log_array(a, b, logsize);
+        fprintf(stderr, "\nfirst %zu bytes of '%s':\n", logsize, bstr);
+        log_array(b, a, logsize);
         exit(1);
     }
 }
 
-static inline bool feq(float a, float b, float epsilon)
-{
-    return fabs(a - b) < epsilon * fmax(1.0, fabs(a));
-}
-
-#define REQUIRE(cond) require((cond), #cond, __FILE__, __LINE__)
-#define RANDOM (rand() / (float) RAND_MAX)
-#define SKIP 77
+#define REQUIRE_MEMEQ(a, b, size) require_memeq(a, b, size, #a, #b, #size, __FILE__, __LINE__)
 
 #define REQUIRE_HANDLE(shmem, type)                                             \
     switch (type) {                                                             \
@@ -82,9 +151,12 @@ static inline bool feq(float a, float b, float epsilon)
         break;                                                                  \
     case PL_HANDLE_WIN32:                                                       \
     case PL_HANDLE_WIN32_KMT:                                                   \
-        REQUIRE(shmem.handle.handle);                                           \
         /* INVALID_HANDLE_VALUE = (-1) */                                       \
         REQUIRE(shmem.handle.handle != (void *)(intptr_t) (-1));                \
+        /* fallthrough */                                                       \
+    case PL_HANDLE_MTL_TEX:                                                     \
+    case PL_HANDLE_IOSURFACE:                                                   \
+        REQUIRE(shmem.handle.handle);                                           \
         break;                                                                  \
     case PL_HANDLE_HOST_PTR:                                                    \
         REQUIRE(shmem.handle.ptr);                                              \
