@@ -37,7 +37,7 @@ struct vk_fun {
 struct vk_ext {
     const char *name;
     uint32_t core_ver;
-    struct vk_fun *funs;
+    const struct vk_fun *funs;
 };
 
 #define PL_VK_INST_FUN(N)                   \
@@ -57,6 +57,7 @@ static const char *vk_instance_extensions[] = {
     VK_EXT_SWAPCHAIN_COLOR_SPACE_EXTENSION_NAME,
     VK_KHR_EXTERNAL_MEMORY_CAPABILITIES_EXTENSION_NAME,
     VK_KHR_EXTERNAL_SEMAPHORE_CAPABILITIES_EXTENSION_NAME,
+    VK_KHR_GET_SURFACE_CAPABILITIES_2_EXTENSION_NAME,
 };
 
 // List of mandatory instance-level function pointers, including functions
@@ -92,7 +93,7 @@ static const struct vk_fun vk_inst_funs[] = {
 static const struct vk_ext vk_device_extensions[] = {
     {
         .name = VK_KHR_SWAPCHAIN_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(AcquireNextImageKHR),
             PL_VK_DEV_FUN(CreateSwapchainKHR),
             PL_VK_DEV_FUN(DestroySwapchainKHR),
@@ -102,46 +103,46 @@ static const struct vk_ext vk_device_extensions[] = {
         },
     }, {
         .name = VK_KHR_PUSH_DESCRIPTOR_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(CmdPushDescriptorSetKHR),
             {0}
         },
     }, {
         .name = VK_KHR_EXTERNAL_MEMORY_FD_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetMemoryFdKHR),
             {0}
         },
     }, {
         .name = VK_EXT_EXTERNAL_MEMORY_DMA_BUF_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetMemoryFdPropertiesKHR),
             {0}
         },
 #ifdef PL_HAVE_WIN32
     }, {
         .name = VK_KHR_EXTERNAL_MEMORY_WIN32_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetMemoryWin32HandleKHR),
             {0}
         },
 #endif
     }, {
         .name = VK_EXT_EXTERNAL_MEMORY_HOST_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetMemoryHostPointerPropertiesEXT),
             {0}
         },
     }, {
         .name = VK_KHR_EXTERNAL_SEMAPHORE_FD_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetSemaphoreFdKHR),
             {0}
         },
 #ifdef PL_HAVE_WIN32
     }, {
         .name = VK_KHR_EXTERNAL_SEMAPHORE_WIN32_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetSemaphoreWin32HandleKHR),
             {0}
         },
@@ -150,14 +151,14 @@ static const struct vk_ext vk_device_extensions[] = {
         .name = VK_EXT_PCI_BUS_INFO_EXTENSION_NAME,
     }, {
         .name = VK_EXT_HDR_METADATA_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(SetHdrMetadataEXT),
             {0}
         },
     }, {
         .name = VK_EXT_HOST_QUERY_RESET_EXTENSION_NAME,
         .core_ver = VK_API_VERSION_1_2,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(ResetQueryPoolEXT),
             {0}
         },
@@ -166,20 +167,36 @@ static const struct vk_ext vk_device_extensions[] = {
         .core_ver = VK_API_VERSION_1_2,
     }, {
         .name = VK_EXT_IMAGE_DRM_FORMAT_MODIFIER_EXTENSION_NAME,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(GetImageDrmFormatModifierPropertiesEXT),
             {0}
         },
     }, {
         .name = VK_KHR_TIMELINE_SEMAPHORE_EXTENSION_NAME,
         .core_ver = VK_API_VERSION_1_2,
-        .funs = (struct vk_fun[]) {
+        .funs = (const struct vk_fun[]) {
             PL_VK_DEV_FUN(WaitSemaphoresKHR),
             {0}
         },
 #ifdef VK_KHR_portability_subset
     }, {
         .name = VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME,
+#endif
+#ifdef VK_EXT_metal_objects
+    }, {
+        .name = VK_EXT_METAL_OBJECTS_EXTENSION_NAME,
+        .funs = (const struct vk_fun[]) {
+            PL_VK_DEV_FUN(ExportMetalObjectsEXT),
+            {0}
+        },
+#endif
+#ifdef VK_EXT_full_screen_exclusive
+    }, {
+        .name = VK_EXT_FULL_SCREEN_EXCLUSIVE_EXTENSION_NAME,
+        .funs = (const struct vk_fun[]) {
+            PL_VK_DEV_FUN(AcquireFullScreenExclusiveModeEXT),
+            {0}
+        },
 #endif
     },
 };
@@ -204,6 +221,12 @@ const char * const pl_vulkan_recommended_extensions[] = {
 #ifdef VK_KHR_portability_subset
     VK_KHR_PORTABILITY_SUBSET_EXTENSION_NAME,
 #endif
+#ifdef VK_EXT_metal_objects
+    VK_EXT_METAL_OBJECTS_EXTENSION_NAME,
+#endif
+#ifdef VK_EXT_full_screen_exclusive
+    VK_EXT_FULL_SCREEN_EXCLUSIVE_EXTENSION_NAME,
+#endif
 };
 
 const int pl_vulkan_num_recommended_extensions =
@@ -215,6 +238,14 @@ static_assert(PL_ARRAY_SIZE(pl_vulkan_recommended_extensions) + 1 ==
               "pl_vulkan_recommended_extensions out of sync with "
               "vk_device_extensions?");
 
+#ifdef VK_VERSION_1_3
+// Needed for LocalSizeId execution mode
+static const VkPhysicalDeviceMaintenance4Features device_maintenance4_features = {
+    .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_MAINTENANCE_4_FEATURES,
+    .maintenance4 = true
+};
+#endif
+
 // pNext chain of features we want enabled
 static const VkPhysicalDeviceTimelineSemaphoreFeatures timeline_semaphores = {
     .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_TIMELINE_SEMAPHORE_FEATURES,
@@ -331,6 +362,7 @@ static const struct vk_fun vk_dev_funs[] = {
     PL_VK_DEV_FUN(InvalidateMappedMemoryRanges),
     PL_VK_DEV_FUN(MapMemory),
     PL_VK_DEV_FUN(QueueSubmit),
+    PL_VK_DEV_FUN(QueueWaitIdle),
     PL_VK_DEV_FUN(ResetEvent),
     PL_VK_DEV_FUN(ResetFences),
     PL_VK_DEV_FUN(SetDebugUtilsObjectNameEXT),
@@ -412,6 +444,7 @@ static VkBool32 VKAPI_PTR vk_dbg_utils_cb(VkDebugUtilsMessageSeverityFlagBitsEXT
     case 0x8928392f: // UNASSIGNED-BestPractices-NonSuccess-Result
     case 0xdc18ad6b: // UNASSIGNED-BestPractices-vkAllocateMemory-small-allocation
     case 0xb3d4346b: // UNASSIGNED-BestPractices-vkBindMemory-small-dedicated-allocation
+    case 0x6cfe18a5: // UNASSIGNED-BestPractices-SemaphoreCount
         return false;
 
     // Work-around for validation layers bug
@@ -526,32 +559,6 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
         },
     };
 
-#ifdef VK_EXT_validation_features
-    // Try enabling as many validation features as possible. Ignored for
-    // instances not supporting VK_EXT_validation_features.
-    VkValidationFeatureEnableEXT validation_features[] = {
-        VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_EXT,
-        VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_RESERVE_BINDING_SLOT_EXT,
-# if VK_EXT_VALIDATION_FEATURES_SPEC_VERSION >= 2
-        VK_VALIDATION_FEATURE_ENABLE_BEST_PRACTICES_EXT,
-# endif
-    };
-
-    VkValidationFeaturesEXT vinfo = {
-        .sType = VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT,
-        .pEnabledValidationFeatures = validation_features,
-        .enabledValidationFeatureCount = PL_ARRAY_SIZE(validation_features),
-    };
-
-    if (params->debug_extra)
-        info.pNext = &vinfo;
-#else
-    if (params->debug_extra) {
-        pl_warn(log, "Enabled extra debugging but vulkan headers too old to "
-                "support it, please update vulkan and recompile libplacebo!");
-    }
-#endif
-
     // Enumerate all supported layers
     PL_VK_LOAD_FUN(NULL, EnumerateInstanceLayerProperties, get_addr);
     uint32_t num_layers_avail = 0;
@@ -576,12 +583,14 @@ pl_vk_inst pl_vk_inst_create(pl_log log, const struct pl_vk_inst_params *params)
     // This layer has to be initialized first, otherwise all sorts of weirdness
     // happens (random segfaults, yum)
     bool debug = params->debug;
+    uint32_t debug_layer = 0; // layer idx of debug layer
     if (debug) {
         for (int i = 0; i < PL_ARRAY_SIZE(debug_layers); i++) {
             for (int n = 0; n < num_layers_avail; n++) {
                 if (strcmp(debug_layers[i], layers_avail[n].layerName) != 0)
                     continue;
 
+                debug_layer = n;
                 pl_info(log, "Enabling debug meta layer: %s", debug_layers[i]);
                 PL_ARRAY_APPEND(tmp, layers, debug_layers[i]);
                 goto debug_layers_done;
@@ -724,14 +733,19 @@ next_opt_user_ext: ;
 
     // If debugging is enabled, load the necessary debug utils extension
     if (debug) {
+        const char * const ext = VK_EXT_DEBUG_UTILS_EXTENSION_NAME;
         for (int n = 0; n < num_exts_avail; n++) {
-            const char * const debug_ext = VK_EXT_DEBUG_UTILS_EXTENSION_NAME;
-            if (strcmp(debug_ext, exts_avail[n].extensionName) != 0)
-                continue;
+            if (strcmp(ext, exts_avail[n].extensionName) == 0) {
+                PL_ARRAY_APPEND(tmp, exts, ext);
+                goto debug_ext_done;
+            }
+        }
 
-            pl_info(log, "Enabling debug report extension: %s", debug_ext);
-            PL_ARRAY_APPEND(tmp, exts, debug_ext);
-            goto debug_ext_done;
+        for (int n = 0; n < layer_exts[debug_layer].num_exts; n++) {
+            if (strcmp(ext, layer_exts[debug_layer].exts[n].extensionName) == 0) {
+                PL_ARRAY_APPEND(tmp, exts, ext);
+                goto debug_ext_done;
+            }
         }
 
         // No extension found
@@ -743,6 +757,44 @@ next_opt_user_ext: ;
 
 debug_ext_done: ;
 
+    if (debug && params->debug_extra) {
+
+        // Try enabling as many validation features as possible
+        static const VkValidationFeatureEnableEXT validation_features[] = {
+            VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_EXT,
+            VK_VALIDATION_FEATURE_ENABLE_GPU_ASSISTED_RESERVE_BINDING_SLOT_EXT,
+            VK_VALIDATION_FEATURE_ENABLE_BEST_PRACTICES_EXT,
+        };
+
+        static const VkValidationFeaturesEXT vinfo = {
+            .sType = VK_STRUCTURE_TYPE_VALIDATION_FEATURES_EXT,
+            .pEnabledValidationFeatures = validation_features,
+            .enabledValidationFeatureCount = PL_ARRAY_SIZE(validation_features),
+        };
+
+        const char * const ext = VK_EXT_VALIDATION_FEATURES_EXTENSION_NAME;
+        for (int n = 0; n < num_exts_avail; n++) {
+            if (strcmp(ext, exts_avail[n].extensionName) == 0) {
+                PL_ARRAY_APPEND(tmp, exts, ext);
+                vk_link_struct(&info, &vinfo);
+                goto debug_extra_ext_done;
+            }
+        }
+
+        for (int n = 0; n < layer_exts[debug_layer].num_exts; n++) {
+            if (strcmp(ext, layer_exts[debug_layer].exts[n].extensionName) == 0) {
+                PL_ARRAY_APPEND(tmp, exts, ext);
+                vk_link_struct(&info, &vinfo);
+                goto debug_extra_ext_done;
+            }
+        }
+
+        pl_warn(log, "GPU-assisted validation enabled but not supported by "
+                "instance, disabling...");
+    }
+
+debug_extra_ext_done: ;
+
     info.ppEnabledExtensionNames = exts.elem;
     info.enabledExtensionCount = exts.num;
     info.ppEnabledLayerNames = layers.elem;
@@ -765,9 +817,9 @@ debug_ext_done: ;
         goto error;
     }
 
-    struct pl_vk_inst *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct priv);
+    struct pl_vk_inst_t *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct priv);
     struct priv *p = PL_PRIV(pl_vk);
-    *pl_vk = (struct pl_vk_inst) {
+    *pl_vk = (struct pl_vk_inst_t) {
         .instance = inst,
         .api_version = api_ver,
         .get_proc_addr = get_addr,
@@ -817,13 +869,15 @@ void pl_vulkan_destroy(pl_vulkan *pl_vk)
         return;
 
     struct vk_ctx *vk = PL_PRIV(*pl_vk);
-    pl_gpu_destroy((*pl_vk)->gpu);
-    vk_malloc_destroy(&vk->ma);
-
     if (vk->dev) {
-        PL_DEBUG(vk, "Waiting for remaining commands...");
-        vk_wait_idle(vk);
-        pl_assert(vk->cmds_pending.num == 0);
+        if ((*pl_vk)->gpu) {
+            PL_DEBUG(vk, "Waiting for remaining commands...");
+            pl_gpu_finish((*pl_vk)->gpu);
+            pl_assert(vk->cmds_pending.num == 0);
+
+            pl_gpu_destroy((*pl_vk)->gpu);
+        }
+        vk_malloc_destroy(&vk->ma);
         for (int i = 0; i < vk->pools.num; i++)
             vk_cmdpool_destroy(vk, vk->pools.elem[i]);
 
@@ -938,7 +992,7 @@ VkPhysicalDevice pl_vulkan_choose_device(pl_log log,
         }
 
         if (!params->allow_software && t == VK_PHYSICAL_DEVICE_TYPE_CPU) {
-            PL_DEBUG(vk, "      -> excluding due to params->allow_software");
+            PL_DEBUG(vk, "      -> excluding due to !params->allow_software");
             continue;
         }
 
@@ -966,7 +1020,7 @@ static int find_qf(VkQueueFamilyProperties *qfs, int qfnum, VkQueueFlags flags)
 {
     int idx = -1;
     for (int i = 0; i < qfnum; i++) {
-        if (!(qfs[i].queueFlags & flags))
+        if ((qfs[i].queueFlags & flags) != flags)
             continue;
 
         // QF is more specialized. Since we don't care about other bits like
@@ -1129,11 +1183,21 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         }
     }
 
+    VkPhysicalDeviceFeatures2 *device_features;
+    device_features = vk_chain_memdup(tmp, &pl_vulkan_recommended_features);
+
+#ifdef VK_VERSION_1_3
+    if (vk->api_ver >= VK_API_VERSION_1_3) {
+        vk_link_struct(device_features,
+                       vk_struct_memdup(tmp, &device_maintenance4_features));
+    }
+#endif
+
     // Query all supported device features by constructing a pNext chain
     // starting with the features we care about and ending with whatever
     // features were requested by the user
     vk->features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2_KHR;
-    for (const VkBaseInStructure *in = pl_vulkan_recommended_features.pNext;
+    for (const VkBaseInStructure *in = device_features->pNext;
             in; in = in->pNext)
         vk_link_struct(&vk->features, vk_struct_memdup(vk->alloc, in));
 
@@ -1168,7 +1232,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
             chain; chain = chain->pNext)
     {
         const VkBaseInStructure *in_a, *in_b;
-        in_a = vk_find_struct(&pl_vulkan_recommended_features, chain->sType);
+        in_a = vk_find_struct(device_features, chain->sType);
         in_b = vk_find_struct(params->features, chain->sType);
         in_a = PL_DEF(in_a, in_b);
         in_b = PL_DEF(in_b, in_a);
@@ -1229,7 +1293,7 @@ static bool device_init(struct vk_ctx *vk, const struct pl_vulkan_params *params
         }
 
         for (int n = 0; n < pool->num_queues; n++)
-            PL_VK_NAME(QUEUE, pool->queues[n], qf_name);
+            PL_VK_NAME_HANDLE(QUEUE, pool->queues[n], qf_name);
     }
 
     pl_free(tmp);
@@ -1266,7 +1330,7 @@ static void unlock_queue(pl_vulkan pl_vk, int qf, int qidx)
     vk->unlock_queue(vk->queue_ctx, qf, qidx);
 }
 
-static bool finalize_context(struct pl_vulkan *pl_vk, int max_glsl_version)
+static bool finalize_context(struct pl_vulkan_t *pl_vk, int max_glsl_version)
 {
     struct vk_ctx *vk = PL_PRIV(pl_vk);
 
@@ -1295,6 +1359,7 @@ static bool finalize_context(struct pl_vulkan *pl_vk, int max_glsl_version)
     pl_vk->instance = vk->inst;
     pl_vk->phys_device = vk->physd;
     pl_vk->device = vk->dev;
+    pl_vk->get_proc_addr = vk->GetInstanceProcAddr;
     pl_vk->api_version = vk->api_ver;
     pl_vk->extensions = vk->exts.elem;
     pl_vk->num_extensions = vk->exts.num;
@@ -1347,7 +1412,7 @@ static bool finalize_context(struct pl_vulkan *pl_vk, int max_glsl_version)
 pl_vulkan pl_vulkan_create(pl_log log, const struct pl_vulkan_params *params)
 {
     params = PL_DEF(params, &pl_vulkan_default_params);
-    struct pl_vulkan *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct vk_ctx);
+    struct pl_vulkan_t *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct vk_ctx);
     struct vk_ctx *vk = PL_PRIV(pl_vk);
     *vk = (struct vk_ctx) {
         .vulkan = pl_vk,
@@ -1456,7 +1521,7 @@ pl_vulkan pl_vulkan_import(pl_log log, const struct pl_vulkan_import_params *par
 {
     void *tmp = pl_tmp(NULL);
 
-    struct pl_vulkan *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct vk_ctx);
+    struct pl_vulkan_t *pl_vk = pl_zalloc_obj(NULL, pl_vk, struct vk_ctx);
     struct vk_ctx *vk = PL_PRIV(pl_vk);
     *vk = (struct vk_ctx) {
         .vulkan = pl_vk,
