@@ -19,63 +19,66 @@
 
 void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
                     VkPipelineStageFlags stage, VkAccessFlags access,
-                    VkImageLayout layout, bool export)
+                    VkImageLayout layout, uint32_t qf)
 {
     struct pl_vk *p = PL_PRIV(gpu);
     struct vk_ctx *vk = p->vk;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
     pl_rc_ref(&tex_vk->rc);
     pl_assert(!tex_vk->held);
-
-    for (int i = 0; i < tex_vk->ext_deps.num; i++)
-        vk_cmd_dep(cmd, stage, tex_vk->ext_deps.elem[i]);
-    tex_vk->ext_deps.num = 0;
-
-    struct vk_sync_scope last;
-    bool is_trans = layout != tex_vk->layout;
-    last = vk_sem_barrier(vk, cmd, &tex_vk->sem, stage, access, is_trans || export);
+    pl_assert(!tex_vk->num_planes);
 
     // CONCURRENT images require transitioning to/from IGNORED, EXCLUSIVE
     // images require transitioning to/from the concrete QF index
-    uint32_t qf = vk->pools.num > 1 ? VK_QUEUE_FAMILY_IGNORED : cmd->pool->qf;
+    if (vk->pools.num == 1) {
+        if (tex_vk->qf == VK_QUEUE_FAMILY_IGNORED)
+            tex_vk->qf = cmd->pool->qf;
+        if (qf == VK_QUEUE_FAMILY_IGNORED)
+            qf = cmd->pool->qf;
+    }
+
+    struct vk_sync_scope last;
+    bool is_trans = layout != tex_vk->layout, is_xfer = qf != tex_vk->qf;
+    last = vk_sem_barrier(vk, cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
+
     VkImageMemoryBarrier barr = {
         .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
         .oldLayout = tex_vk->layout,
         .newLayout = layout,
-        .srcQueueFamilyIndex = qf,
-        .dstQueueFamilyIndex = export ? VK_QUEUE_FAMILY_EXTERNAL_KHR : qf,
+        .srcQueueFamilyIndex = tex_vk->qf,
+        .dstQueueFamilyIndex = qf,
         .srcAccessMask = last.access,
         .dstAccessMask = access,
         .image = tex_vk->img,
         .subresourceRange = {
-            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+            .aspectMask = tex_vk->aspect,
             .levelCount = 1,
             .layerCount = 1,
         },
     };
 
-    if (tex_vk->ext_sync) {
-        if (tex_vk->layout != VK_IMAGE_LAYOUT_UNDEFINED) {
-            barr.srcQueueFamilyIndex = VK_QUEUE_FAMILY_EXTERNAL_KHR;
-            pl_assert(!export); // can't re-export exported images
-        }
-        vk_cmd_callback(cmd, (vk_cb) vk_sync_deref, gpu, tex_vk->ext_sync);
-        tex_vk->ext_sync = NULL;
-    }
-
     if (tex_vk->may_invalidate) {
         tex_vk->may_invalidate = false;
         barr.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
     }
 
-    bool is_xfer = barr.srcQueueFamilyIndex != barr.dstQueueFamilyIndex;
     if (last.access || is_trans || is_xfer) {
         vk->CmdPipelineBarrier(cmd->buf, last.stage, stage, 0, 0, NULL,
                                0, NULL, 1, &barr);
     }
 
+    tex_vk->qf = qf;
     tex_vk->layout = layout;
     vk_cmd_callback(cmd, (vk_cb) vk_tex_deref, gpu, tex);
+
+    for (int i = 0; i < tex_vk->ext_deps.num; i++)
+        vk_cmd_dep(cmd, stage, tex_vk->ext_deps.elem[i]);
+    tex_vk->ext_deps.num = 0;
+
+    if (tex_vk->ext_sync) {
+        vk_cmd_callback(cmd, (vk_cb) vk_sync_deref, gpu, tex_vk->ext_sync);
+        tex_vk->ext_sync = NULL;
+    }
 }
 
 static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
@@ -88,9 +91,10 @@ static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
     vk_sync_deref(gpu, tex_vk->ext_sync);
-    vk_sem_uninit(vk, &tex_vk->sem);
     vk->DestroyFramebuffer(vk->dev, tex_vk->framebuffer, PL_VK_ALLOC);
     vk->DestroyImageView(vk->dev, tex_vk->view, PL_VK_ALLOC);
+    for (int i = 0; i < tex_vk->num_planes; i++)
+        vk_tex_deref(gpu, tex->planes[i]);
     if (!tex_vk->external_img) {
         vk->DestroyImage(vk->dev, tex_vk->img, PL_VK_ALLOC);
         vk_malloc_free(vk->ma, &tex_vk->mem);
@@ -120,12 +124,13 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
     pl_assert(tex_vk->img);
     PL_VK_NAME(IMAGE, tex_vk->img, debug_tag);
-
     pl_rc_init(&tex_vk->rc);
-    if (!vk_sem_init(vk, &tex_vk->sem, debug_tag))
-        return false;
+    if (tex_vk->num_planes)
+        return true;
+    vk_sem_init(&tex_vk->sem);
     tex_vk->layout = VK_IMAGE_LAYOUT_UNDEFINED;
     tex_vk->transfer_queue = GRAPHICS;
+    tex_vk->qf = VK_QUEUE_FAMILY_IGNORED; // will be set on first use, if needed
 
     // Always use the transfer pool if available, for efficiency
     if ((params->host_writable || params->host_readable) && vk->pool_transfer)
@@ -146,13 +151,13 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
             [VK_IMAGE_TYPE_3D] = VK_IMAGE_VIEW_TYPE_3D,
         };
 
-        VkImageViewCreateInfo vinfo = {
+        const VkImageViewCreateInfo vinfo = {
             .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
             .image = tex_vk->img,
             .viewType = viewType[tex_vk->type],
             .format = tex_vk->img_fmt,
             .subresourceRange = {
-                .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+                .aspectMask = tex_vk->aspect,
                 .levelCount = 1,
                 .layerCount = 1,
             },
@@ -232,13 +237,18 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     VkExternalMemoryHandleTypeFlagBitsKHR vk_handle_type = vk_mem_handle_type(handle_type);
 
     struct pl_tex_t *tex = pl_zalloc_obj(NULL, tex, struct pl_tex_vk);
+    pl_fmt fmt = params->format;
     tex->params = *params;
     tex->params.initial_data = NULL;
     tex->sampler_type = PL_SAMPLER_NORMAL;
 
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
-    struct pl_fmt_vk *fmtp = PL_PRIV(params->format);
+    struct pl_fmt_vk *fmtp = PL_PRIV(fmt);
     tex_vk->img_fmt = fmtp->vk_fmt->tfmt;
+    tex_vk->num_planes = fmt->num_planes;
+    for (int i = 0; i < tex_vk->num_planes; i++)
+        tex_vk->aspect |= VK_IMAGE_ASPECT_PLANE_0_BIT << i;
+    tex_vk->aspect = PL_DEF(tex_vk->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
 
     switch (pl_tex_params_dimension(*params)) {
     case 1: tex_vk->type = VK_IMAGE_TYPE_1D; break;
@@ -246,9 +256,9 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     case 3: tex_vk->type = VK_IMAGE_TYPE_3D; break;
     }
 
-    if (params->format->emulated) {
-        tex_vk->texel_fmt = pl_find_fmt(gpu, params->format->type, 1, 0,
-                                        params->format->host_bits[0],
+    if (fmt->emulated) {
+        tex_vk->texel_fmt = pl_find_fmt(gpu, fmt->type, 1, 0,
+                                        fmt->host_bits[0],
                                         PL_FMT_CAP_TEXEL_UNIFORM);
         if (!tex_vk->texel_fmt) {
             PL_ERR(gpu, "Failed picking texel format for emulated texture!");
@@ -260,7 +270,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         // based on the size of pl_tex_transfer_params.row_pitch, but for now
         // this should be enough.
         uint64_t texels = params->w * PL_DEF(params->h, 1) * PL_DEF(params->d, 1) *
-                          params->format->num_components;
+                          fmt->num_components;
 
         if (texels > gpu->limits.max_buffer_texels) {
             PL_ERR(gpu, "Failed creating texture with emulated texture format: "
@@ -277,11 +287,16 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
 
     if (fmtp->blit_emulated) {
         // Enable what's required for sampling
-        tex->params.sampleable = params->format->caps & PL_FMT_CAP_SAMPLEABLE;
+        tex->params.sampleable = fmt->caps & PL_FMT_CAP_SAMPLEABLE;
         tex->params.storable = true;
     }
 
+    // Blit emulation on planar textures requires storage
+    if ((params->blit_src || params->blit_dst) && tex_vk->num_planes)
+        tex->params.storable = true;
+
     VkImageUsageFlags usage = 0;
+    VkImageCreateFlags flags = 0;
     if (tex->params.sampleable)
         usage |= VK_IMAGE_USAGE_SAMPLED_BIT;
     if (tex->params.renderable)
@@ -300,6 +315,11 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         usage = VK_IMAGE_USAGE_TRANSFER_DST_BIT;
     }
 
+    if (tex_vk->num_planes) {
+        flags |= VK_IMAGE_CREATE_MUTABLE_FORMAT_BIT |
+                 VK_IMAGE_CREATE_EXTENDED_USAGE_BIT;
+    }
+
     // FIXME: Since we can't keep track of queue family ownership properly,
     // and we don't know in advance what types of queue families this image
     // will belong to, we're forced to share all of our images between all
@@ -332,8 +352,8 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
 
     VkImageDrmFormatModifierListCreateInfoEXT drm_list = {
         .sType = VK_STRUCTURE_TYPE_IMAGE_DRM_FORMAT_MODIFIER_LIST_CREATE_INFO_EXT,
-        .drmFormatModifierCount = params->format->num_modifiers,
-        .pDrmFormatModifiers = params->format->modifiers,
+        .drmFormatModifierCount = fmt->num_modifiers,
+        .pDrmFormatModifiers = fmt->modifiers,
     };
 
     VkExternalMemoryImageCreateInfoKHR ext_info = {
@@ -356,6 +376,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         .samples = VK_SAMPLE_COUNT_1_BIT,
         .tiling = VK_IMAGE_TILING_OPTIMAL,
         .usage = usage,
+        .flags = flags,
         .initialLayout = VK_IMAGE_LAYOUT_UNDEFINED,
         .sharingMode = vk->pools.num > 1 ? VK_SHARING_MODE_CONCURRENT
                                          : VK_SHARING_MODE_EXCLUSIVE,
@@ -368,6 +389,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         .export_handle = params->export_handle,
         .import_handle = params->import_handle,
         .shared_mem = params->shared_mem,
+        .debug_tag = params->debug_tag,
     };
 
     if (params->import_handle == PL_HANDLE_DMA_BUF) {
@@ -432,6 +454,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     VkResult res;
     res = vk->GetPhysicalDeviceImageFormatProperties2KHR(vk->physd, &pinfo, &props);
     if (res == VK_ERROR_FORMAT_NOT_SUPPORTED) {
+        PL_DEBUG(gpu, "Texture creation failed: not supported");
         goto error;
     } else {
         PL_VK_ASSERT(res, "Querying image format properties");
@@ -495,6 +518,45 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         VK(vk->BindImageMemory(vk->dev, tex_vk->img, mem->vkmem, mem->offset));
     }
 
+    static const char * const plane_names[4] = {
+        "plane 0", "plane 1", "plane 2", "plane 3",
+    };
+
+    if (tex_vk->num_planes) {
+        for (int i = 0; i < tex_vk->num_planes; i++) {
+            struct pl_tex_t *plane;
+
+            pl_assert(tex_vk->type == VK_IMAGE_TYPE_2D);
+            plane = (struct pl_tex_t *) pl_vulkan_wrap(gpu, pl_vulkan_wrap_params(
+                .image      = tex_vk->img,
+                .aspect     = VK_IMAGE_ASPECT_PLANE_0_BIT << i,
+                .width      = PL_RSHIFT_UP(tex->params.w, fmt->planes[i].shift_x),
+                .height     = PL_RSHIFT_UP(tex->params.h, fmt->planes[i].shift_y),
+                .format     = fmtp->vk_fmt->pfmt[i].fmt,
+                .usage      = usage,
+                .user_data  = params->user_data,
+                .debug_tag  = PL_DEF(params->debug_tag, plane_names[i]),
+            ));
+            if (!plane)
+                goto error;
+            plane->parent = tex;
+            tex->planes[i] = plane;
+            tex_vk->planes[i] = PL_PRIV(plane);
+            tex_vk->planes[i]->held = false;
+            tex_vk->planes[i]->layout = tex_vk->layout;
+        }
+
+        // Explicitly mask out all usage flags from planar parent images
+        pl_assert(!fmt->caps);
+        tex->params.sampleable      = false;
+        tex->params.renderable      = false;
+        tex->params.storable        = false;
+        tex->params.blit_src        = false;
+        tex->params.blit_dst        = false;
+        tex->params.host_writable   = false;
+        tex->params.host_readable   = false;
+    }
+
     if (!vk_init_image(gpu, tex, debug_tag))
         goto error;
 
@@ -563,6 +625,35 @@ void vk_tex_invalidate(pl_gpu gpu, pl_tex tex)
 {
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
     tex_vk->may_invalidate = true;
+    for (int i = 0; i < tex_vk->num_planes; i++)
+        tex_vk->planes[i]->may_invalidate = true;
+}
+
+static bool tex_clear_fallback(pl_gpu gpu, pl_tex tex,
+                               const union pl_clear_color color)
+{
+    pl_tex pixel = pl_tex_create(gpu, pl_tex_params(
+        .w = 1,
+        .h = 1,
+        .format = tex->params.format,
+        .storable = true,
+        .blit_src = true,
+        .blit_dst = true,
+    ));
+    if (!pixel)
+        return false;
+
+    pl_tex_clear_ex(gpu, pixel, color);
+
+    pl_assert(tex->params.storable);
+    pl_tex_blit(gpu, pl_tex_blit_params(
+        .src = pixel,
+        .dst = tex,
+        .sample_mode = PL_TEX_SAMPLE_NEAREST,
+    ));
+
+    pl_tex_destroy(gpu, &pixel);
+    return true;
 }
 
 void vk_tex_clear_ex(pl_gpu gpu, pl_tex tex, const union pl_clear_color color)
@@ -571,6 +662,15 @@ void vk_tex_clear_ex(pl_gpu gpu, pl_tex tex, const union pl_clear_color color)
     struct vk_ctx *vk = p->vk;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
+    if (tex_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT) {
+        if (!tex_clear_fallback(gpu, tex,  color)) {
+            PL_ERR(gpu, "Failed clearing imported planar image: color aspect "
+                   "clears disallowed by spec and no shader fallback "
+                   "available");
+        }
+        return;
+    }
+
     struct vk_cmd *cmd = CMD_BEGIN(GRAPHICS);
     if (!cmd)
         return;
@@ -578,11 +678,12 @@ void vk_tex_clear_ex(pl_gpu gpu, pl_tex tex, const union pl_clear_color color)
     vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
                    VK_ACCESS_TRANSFER_WRITE_BIT,
                    VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                   false);
+                   VK_QUEUE_FAMILY_IGNORED);
 
     pl_static_assert(sizeof(VkClearColorValue) == sizeof(union pl_clear_color));
     const VkClearColorValue *clearColor = (const VkClearColorValue *) &color;
 
+    pl_assert(tex_vk->aspect == VK_IMAGE_ASPECT_COLOR_BIT);
     static const VkImageSubresourceRange range = {
         .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
         .levelCount = 1,
@@ -604,11 +705,13 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     struct pl_fmt_vk *src_fmtp = PL_PRIV(params->src->params.format);
     struct pl_fmt_vk *dst_fmtp = PL_PRIV(params->dst->params.format);
     bool blit_emulated = src_fmtp->blit_emulated || dst_fmtp->blit_emulated;
+    bool planar_fallback = src_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT ||
+                           dst_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT;
 
     struct pl_rect3d src_rc = params->src_rc, dst_rc = params->dst_rc;
     bool requires_scaling = !pl_rect3d_eq(src_rc, dst_rc);
-    if (requires_scaling && blit_emulated) {
-        if (!pl_tex_blit_compute(gpu, p->dp, params))
+    if ((requires_scaling && blit_emulated) || planar_fallback) {
+        if (!pl_tex_blit_compute(gpu, params))
             PL_ERR(gpu, "Failed emulating texture blit, incompatible textures?");
         return;
     }
@@ -620,17 +723,12 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_TRANSFER_BIT,
                    VK_ACCESS_TRANSFER_READ_BIT,
                    VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
-                   false);
+                   VK_QUEUE_FAMILY_IGNORED);
 
     vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_TRANSFER_BIT,
                    VK_ACCESS_TRANSFER_WRITE_BIT,
                    VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                   false);
-
-    static const VkImageSubresourceLayers layers = {
-        .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
-        .layerCount = 1,
-    };
+                   VK_QUEUE_FAMILY_IGNORED);
 
     // When the blit operation doesn't require scaling, we can use the more
     // efficient vkCmdCopyImage instead of vkCmdBlitImage
@@ -638,8 +736,14 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
         pl_rect3d_normalize(&src_rc);
 
         VkImageCopy region = {
-            .srcSubresource = layers,
-            .dstSubresource = layers,
+            .srcSubresource = {
+                .aspectMask = src_vk->aspect,
+                .layerCount = 1,
+            },
+            .dstSubresource = {
+                .aspectMask = dst_vk->aspect,
+                .layerCount = 1,
+            },
             .srcOffset = {src_rc.x0, src_rc.y0, src_rc.z0},
             .dstOffset = {src_rc.x0, src_rc.y0, src_rc.z0},
             .extent = {
@@ -653,8 +757,14 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
                          dst_vk->img, dst_vk->layout, 1, &region);
     } else {
         VkImageBlit region = {
-            .srcSubresource = layers,
-            .dstSubresource = layers,
+            .srcSubresource = {
+                .aspectMask = src_vk->aspect,
+                .layerCount = 1,
+            },
+            .dstSubresource = {
+                .aspectMask = dst_vk->aspect,
+                .layerCount = 1,
+            },
             .srcOffsets = {{src_rc.x0, src_rc.y0, src_rc.z0},
                            {src_rc.x1, src_rc.y1, src_rc.z1}},
             .dstOffsets = {{dst_rc.x0, dst_rc.y0, dst_rc.z0},
@@ -809,7 +919,7 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         fixed.buf = tbuf;
         fixed.buf_offset = 0;
 
-        bool ok = fmt->emulated ? pl_tex_upload_texel(gpu, p->dp, &fixed)
+        bool ok = fmt->emulated ? pl_tex_upload_texel(gpu, &fixed)
                                 : pl_tex_upload(gpu, &fixed);
 
         pl_buf_destroy(gpu, &tbuf);
@@ -818,14 +928,14 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     } else {
 
         pl_assert(fmt->texel_align == fmt->texel_size);
-        VkBufferImageCopy region = {
+        const VkBufferImageCopy region = {
             .bufferOffset = buf_offset,
             .bufferRowLength = params->row_pitch / fmt->texel_size,
             .bufferImageHeight = params->depth_pitch / params->row_pitch,
             .imageOffset = { rc.x0, rc.y0, rc.z0 },
             .imageExtent = { rc.x1, rc.y1, rc.z1 },
             .imageSubresource = {
-                .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+                .aspectMask = tex_vk->aspect,
                 .layerCount = 1,
             },
         };
@@ -840,7 +950,8 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
                        false);
         vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
                        VK_ACCESS_TRANSFER_WRITE_BIT,
-                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, false);
+                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyBufferToImage(cmd->buf, buf_vk->mem.buf, tex_vk->img,
                                  tex_vk->layout, 1, &region);
 
@@ -896,7 +1007,7 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         fixed.buf = tbuf;
         fixed.buf_offset = 0;
 
-        bool ok = fmt->emulated ? pl_tex_download_texel(gpu, p->dp, &fixed)
+        bool ok = fmt->emulated ? pl_tex_download_texel(gpu, &fixed)
                                 : pl_tex_download(gpu, &fixed);
         if (!ok) {
             pl_buf_destroy(gpu, &tbuf);
@@ -935,14 +1046,14 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     } else {
 
         pl_assert(fmt->texel_align == fmt->texel_size);
-        VkBufferImageCopy region = {
+        const VkBufferImageCopy region = {
             .bufferOffset = buf_offset,
             .bufferRowLength = params->row_pitch / fmt->texel_size,
             .bufferImageHeight = params->depth_pitch / params->row_pitch,
             .imageOffset = { rc.x0, rc.y0, rc.z0 },
             .imageExtent = { rc.x1, rc.y1, rc.z1 },
             .imageSubresource = {
-                .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+                .aspectMask = tex_vk->aspect,
                 .layerCount = 1,
             },
         };
@@ -958,7 +1069,8 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
                        false);
         vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
                        VK_ACCESS_TRANSFER_READ_BIT,
-                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL, false);
+                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyImageToBuffer(cmd->buf, tex_vk->img, tex_vk->layout,
                                  buf_vk->mem.buf, 1, &region);
         vk_buf_flush(gpu, cmd, buf, params->buf_offset, size);
@@ -985,14 +1097,23 @@ bool vk_tex_poll(pl_gpu gpu, pl_tex tex, uint64_t timeout)
     // Opportunistically check if we can re-use this texture without flush
     vk_poll_commands(vk, 0);
     if (pl_rc_count(&tex_vk->rc) == 1)
-        return false;
+        goto skip_blocking;
 
     // Otherwise, we're force to submit any queued command so that the user is
     // guaranteed to see progress eventually, even if they call this in a loop
     CMD_SUBMIT(NULL);
     vk_poll_commands(vk, timeout);
+    if (pl_rc_count(&tex_vk->rc) > 1)
+        return true;
 
-    return pl_rc_count(&tex_vk->rc) > 1;
+    // fall through
+skip_blocking:
+    for (int i = 0; i < tex_vk->num_planes; i++) {
+        if (vk_tex_poll(gpu, tex->planes[i], timeout))
+            return true;
+    }
+
+    return false;
 }
 
 bool vk_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
@@ -1000,12 +1121,18 @@ bool vk_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
     struct pl_sync_vk *sync_vk = PL_PRIV(sync);
 
+    if (tex_vk->num_planes) {
+        PL_ERR(gpu, "`pl_tex_export` cannot be called on planar textures."
+               "Please see `pl_vulkan_hold_ex` for a replacement.");
+        return false;
+    }
+
     struct vk_cmd *cmd = CMD_BEGIN(ANY);
     if (!cmd)
         goto error;
 
     vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                   0, VK_IMAGE_LAYOUT_GENERAL, true);
+                   0, VK_IMAGE_LAYOUT_GENERAL, VK_QUEUE_FAMILY_EXTERNAL);
 
     // Make the next barrier appear as though coming from a different queue
     tex_vk->sem.write.queue = tex_vk->sem.read.queue = NULL;
@@ -1018,6 +1145,7 @@ bool vk_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
     PL_ARRAY_APPEND(tex, tex_vk->ext_deps, (pl_vulkan_sem){ sync_vk->signal });
     pl_rc_ref(&sync_vk->rc);
     tex_vk->ext_sync = sync;
+    tex_vk->qf = VK_QUEUE_FAMILY_EXTERNAL;
     return true;
 
 error:
@@ -1027,69 +1155,119 @@ error:
 
 pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
 {
-    pl_fmt format = NULL;
+    pl_fmt fmt = NULL;
     for (int i = 0; i < gpu->num_formats; i++) {
-        const struct vk_format **fmt = PL_PRIV(gpu->formats[i]);
-        if ((*fmt)->tfmt == params->format) {
-            format = gpu->formats[i];
+        const struct vk_format **vkfmt = PL_PRIV(gpu->formats[i]);
+        if ((*vkfmt)->tfmt == params->format) {
+            fmt = gpu->formats[i];
             break;
         }
     }
 
-    if (!format) {
+    if (!fmt) {
         PL_ERR(gpu, "Could not find pl_fmt suitable for wrapped image "
                "with format %s", vk_fmt_name(params->format));
         return NULL;
     }
 
+    VkImageUsageFlags usage = params->usage;
+    if (fmt->num_planes)
+        usage = 0; // mask capabilities from the base texture
+
     struct pl_tex_t *tex = pl_zalloc_obj(NULL, tex, struct pl_tex_vk);
     tex->params = (struct pl_tex_params) {
-        .format = format,
-        .w = params->width,
-        .h = params->height,
-        .d = params->depth,
-        .sampleable = !!(params->usage & VK_IMAGE_USAGE_SAMPLED_BIT),
-        .renderable = !!(params->usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT),
-        .storable   = !!(params->usage & VK_IMAGE_USAGE_STORAGE_BIT),
-        .blit_src   = !!(params->usage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
-        .blit_dst   = !!(params->usage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
-        .host_writable = !!(params->usage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
-        .host_readable = !!(params->usage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
-        .user_data  = params->user_data,
-        .debug_tag  = params->debug_tag,
+        .format         = fmt,
+        .w              = params->width,
+        .h              = params->height,
+        .d              = params->depth,
+        .sampleable     = !!(usage & VK_IMAGE_USAGE_SAMPLED_BIT),
+        .renderable     = !!(usage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT),
+        .storable       = !!(usage & VK_IMAGE_USAGE_STORAGE_BIT),
+        .blit_src       = !!(usage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
+        .blit_dst       = !!(usage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
+        .host_writable  = !!(usage & VK_IMAGE_USAGE_TRANSFER_DST_BIT),
+        .host_readable  = !!(usage & VK_IMAGE_USAGE_TRANSFER_SRC_BIT),
+        .user_data      = params->user_data,
+        .debug_tag      = params->debug_tag,
     };
 
     // Mask out capabilities not permitted by the `pl_fmt`
 #define MASK(field, cap)                                                        \
     do {                                                                        \
-        if (tex->params.field && !(format->caps & cap)) {                       \
+        if (tex->params.field && !(fmt->caps & cap)) {                          \
             PL_WARN(gpu, "Masking `" #field "` from wrapped texture because "   \
                     "the corresponding format '%s' does not support " #cap,     \
-                    format->name);                                              \
+                    fmt->name);                                                 \
             tex->params.field = false;                                          \
         }                                                                       \
     } while (0)
 
-    MASK(sampleable, PL_FMT_CAP_SAMPLEABLE);
-    MASK(storable,   PL_FMT_CAP_STORABLE);
-    MASK(blit_src,   PL_FMT_CAP_BLITTABLE);
-    MASK(blit_dst,   PL_FMT_CAP_BLITTABLE);
+    MASK(sampleable,    PL_FMT_CAP_SAMPLEABLE);
+    MASK(renderable,    PL_FMT_CAP_RENDERABLE);
+    MASK(storable,      PL_FMT_CAP_STORABLE);
+    MASK(blit_src,      PL_FMT_CAP_BLITTABLE);
+    MASK(blit_dst,      PL_FMT_CAP_BLITTABLE);
+    MASK(host_readable, PL_FMT_CAP_HOST_READABLE);
 #undef MASK
 
     // For simplicity, explicitly mask out blit emulation for wrapped textures
-    struct pl_fmt_vk *fmtp = PL_PRIV(format);
+    struct pl_fmt_vk *fmtp = PL_PRIV(fmt);
     if (fmtp->blit_emulated) {
         tex->params.blit_src = false;
         tex->params.blit_dst = false;
     }
 
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
-    tex_vk->type = VK_IMAGE_TYPE_2D;
+    switch (pl_tex_params_dimension(tex->params)) {
+    case 1: tex_vk->type = VK_IMAGE_TYPE_1D; break;
+    case 2: tex_vk->type = VK_IMAGE_TYPE_2D; break;
+    case 3: tex_vk->type = VK_IMAGE_TYPE_3D; break;
+    }
     tex_vk->external_img = true;
-    tex_vk->held = true;
+    tex_vk->held = !fmt->num_planes;
     tex_vk->img = params->image;
     tex_vk->img_fmt = params->format;
-    tex_vk->usage_flags = params->usage;
+    tex_vk->num_planes = fmt->num_planes;
+    tex_vk->usage_flags = usage;
+    tex_vk->aspect = PL_DEF(params->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
+
+    // Blitting to planar images requires fallback via compute shaders
+    if (tex_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT) {
+        tex->params.blit_src &= tex->params.storable;
+        tex->params.blit_dst &= tex->params.storable;
+    }
+
+    static const char * const wrapped_plane_names[4] = {
+        "wrapped plane 0", "wrapped plane 1", "wrapped plane 2", "wrapped plane 3",
+    };
+
+    for (int i = 0; i < tex_vk->num_planes; i++) {
+        struct pl_tex_t *plane;
+        VkImageAspectFlags aspect = VK_IMAGE_ASPECT_PLANE_0_BIT << i;
+        if (!(aspect & params->aspect)) {
+            PL_INFO(gpu, "Not wrapping plane %d due to aspect bit 0x%x not "
+                    "being contained in supplied params->aspect 0x%x!",
+                    i, (unsigned) aspect, (unsigned) params->aspect);
+            continue;
+        }
+
+        pl_assert(tex_vk->type == VK_IMAGE_TYPE_2D);
+        plane = (struct pl_tex_t *) pl_vulkan_wrap(gpu, pl_vulkan_wrap_params(
+            .image      = tex_vk->img,
+            .aspect     = aspect,
+            .width      = PL_RSHIFT_UP(tex->params.w, fmt->planes[i].shift_x),
+            .height     = PL_RSHIFT_UP(tex->params.h, fmt->planes[i].shift_y),
+            .format     = fmtp->vk_fmt->pfmt[i].fmt,
+            .usage      = params->usage,
+            .user_data  = params->user_data,
+            .debug_tag  = PL_DEF(params->debug_tag, wrapped_plane_names[i]),
+        ));
+        if (!plane)
+            goto error;
+        plane->parent = tex;
+        tex->planes[i] = plane;
+        tex_vk->planes[i] = PL_PRIV(plane);
+    }
 
     if (!vk_init_image(gpu, tex, PL_DEF(params->debug_tag, "wrapped")))
         goto error;
@@ -1114,13 +1292,16 @@ VkImage pl_vulkan_unwrap(pl_gpu gpu, pl_tex tex, VkFormat *out_format,
     return tex_vk->img;
 }
 
-bool pl_vulkan_hold(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
-                    pl_vulkan_sem sem_out)
+bool pl_vulkan_hold_ex(pl_gpu gpu, const struct pl_vulkan_hold_params *params)
 {
-    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
-    pl_assert(sem_out.sem);
+    struct pl_tex_vk *tex_vk = PL_PRIV(params->tex);
+    pl_assert(params->semaphore.sem);
 
-    if (tex_vk->held) {
+    bool held = tex_vk->held;
+    for (int i = 0; i < tex_vk->num_planes; i++)
+        held |= tex_vk->planes[i]->held;
+
+    if (held) {
         PL_ERR(gpu, "Attempting to hold an already held image!");
         return false;
     }
@@ -1131,45 +1312,74 @@ bool pl_vulkan_hold(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
         return false;
     }
 
-    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                   0, layout, false);
-
-    vk_cmd_sig(cmd, sem_out);
-
-    tex_vk->sem.write.queue = tex_vk->sem.read.queue = NULL;
-    tex_vk->held = CMD_SUBMIT(&cmd);
-    return tex_vk->held;
-}
-
-bool pl_vulkan_hold_raw(pl_gpu gpu, pl_tex tex, VkImageLayout *layout,
-                        pl_vulkan_sem sem_out)
-{
-    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
-    bool user_may_invalidate = tex_vk->may_invalidate;
-    if (!pl_vulkan_hold(gpu, tex, tex_vk->layout, sem_out))
-        return false;
-
-    if (user_may_invalidate) {
-        *layout = VK_IMAGE_LAYOUT_UNDEFINED;
-    } else {
-        *layout = tex_vk->layout;
+    VkImageLayout layout = params->layout;
+    if (params->out_layout) {
+        // For planar images, arbitrarily pick the current image layout of the
+        // first plane. This should be fine in practice, since all planes will
+        // share the same usage capabilities.
+        if (tex_vk->num_planes) {
+            layout = tex_vk->planes[0]->layout;
+        } else {
+            layout = tex_vk->layout;
+        }
     }
 
-    return true;
+    bool may_invalidate = true;
+    if (!tex_vk->num_planes) {
+        may_invalidate &= tex_vk->may_invalidate;
+        vk_tex_barrier(gpu, cmd, params->tex,
+                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+                       0, layout, params->qf);
+    }
+
+    for (int i = 0; i < tex_vk->num_planes; i++) {
+        may_invalidate &= tex_vk->planes[i]->may_invalidate;
+        vk_tex_barrier(gpu, cmd, params->tex->planes[i],
+                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+                       0, layout, params->qf);
+    }
+
+    vk_cmd_sig(cmd, params->semaphore);
+    bool ok = CMD_SUBMIT(&cmd);
+
+    if (!tex_vk->num_planes) {
+        tex_vk->sem.write.queue = tex_vk->sem.read.queue = NULL;
+        tex_vk->held = ok;
+    }
+
+    for (int i = 0; i < tex_vk->num_planes; i++) {
+        struct pl_tex_vk *plane_vk = tex_vk->planes[i];
+        plane_vk->sem.write.queue = plane_vk->sem.read.queue = NULL;
+        plane_vk->held = ok;
+    }
+
+    if (ok && params->out_layout)
+        *params->out_layout = may_invalidate ? VK_IMAGE_LAYOUT_UNDEFINED : layout;
+
+    return ok;
 }
 
-void pl_vulkan_release(pl_gpu gpu, pl_tex tex, VkImageLayout layout,
-                       pl_vulkan_sem sem_in)
+void pl_vulkan_release_ex(pl_gpu gpu, const struct pl_vulkan_release_params *params)
 {
-    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    struct pl_tex_vk *tex_vk = PL_PRIV(params->tex);
+    if (tex_vk->num_planes) {
+        struct pl_vulkan_release_params plane_pars = *params;
+        for (int i = 0; i < tex_vk->num_planes; i++) {
+            plane_pars.tex = params->tex->planes[i];
+            pl_vulkan_release_ex(gpu, &plane_pars);
+        }
+        return;
+    }
+
     if (!tex_vk->held) {
         PL_ERR(gpu, "Attempting to release an unheld image?");
         return;
     }
 
-    if (sem_in.sem)
-        PL_ARRAY_APPEND(tex, tex_vk->ext_deps, sem_in);
+    if (params->semaphore.sem)
+        PL_ARRAY_APPEND(params->tex, tex_vk->ext_deps, params->semaphore);
 
-    tex_vk->layout = layout;
+    tex_vk->qf = params->qf;
+    tex_vk->layout = params->layout;
     tex_vk->held = false;
 }
